import autolens as al
import numpy as np
import pytest

import autofit as af
from test.unit.mock.model.mock_profiles import MockLightProfile


class TestFitProperties:
    def test__total_inversions(self, lens_uv_plane_data_7):

        g0 = al.Galaxy(redshift=0.5)

        g1 = al.Galaxy(redshift=1.0)

        g2 = al.Galaxy(redshift=2.0)

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

        fit = al.LensUVPlaneFit.from_lens_data_and_tracer(
            lens_data=lens_uv_plane_data_7, tracer=tracer
        )

        assert fit.total_inversions == 0

        g2 = al.Galaxy(
            redshift=2.0,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

        fit = al.LensUVPlaneFit.from_lens_data_and_tracer(
            lens_data=lens_uv_plane_data_7, tracer=tracer
        )

        assert fit.total_inversions == 1

        g0 = al.Galaxy(
            redshift=0.5,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        g1 = al.Galaxy(
            redshift=1.0,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        g2 = al.Galaxy(
            redshift=2.0,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

        fit = al.LensUVPlaneFit.from_lens_data_and_tracer(
            lens_data=lens_uv_plane_data_7, tracer=tracer
        )

        assert fit.total_inversions == 3


class TestLikelihood:
    def test__1x2_image__1x2_visibilities__simple_fourier_transform(self):
        # The image plane image generated by the galaxy is [1.0, 1.0]

        # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

        uv_wavelengths = np.array([[0.0, 0.0]])

        uv_plane_data = al.UVPlaneData(
            shape=(1, 2),
            visibilities=5.0 * np.ones((1, 2)),
            pixel_scale=1.0,
            noise_map=np.ones((1,)),
            uv_wavelengths=uv_wavelengths,
            primary_beam=None,
        )
        uv_plane_data.visibilities[0, 1] = 4.0

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_uv_plane_data_7 = al.LensUVPlaneData(
            uv_plane_data=uv_plane_data, mask=mask
        )

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2))
        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        fit = al.LensUVPlaneFit.from_lens_data_and_tracer(
            lens_data=lens_uv_plane_data_7, tracer=tracer
        )

        assert (fit._mask == np.array([False, False])).all()

        assert (fit._data == np.array([5.0, 4.0])).all()
        assert (fit.visibilities() == np.array([[5.0, 4.0]])).all()
        assert (fit._noise_map == np.array([[1.0, 1.0]])).all()
        assert (fit.noise_map() == np.array([[1.0, 1.0]])).all()
        assert (fit._model_data == np.array([2.0, 0.0])).all()
        assert (fit.model_visibilities() == np.array([2.0, 0.0])).all()
        assert (fit._residual_map == np.array([3.0, 4.0])).all()
        assert (fit.residual_map() == np.array([3.0, 4.0])).all()
        assert (fit._normalized_residual_map == np.array([3.0, 4.0])).all()
        assert (fit.normalized_residual_map() == np.array([3.0, 4.0])).all()
        assert (fit._chi_squared_map == np.array([9.0, 16.0])).all()
        assert (fit.chi_squared_map() == np.array([9.0, 16.0])).all()

        assert fit.chi_squared == 25.0
        assert fit.reduced_chi_squared == 25.0 / 2.0
        assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
        assert fit.likelihood == -0.5 * (25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0))

    def test__3x2_image__2x2_visibilities__use_transformer_for_likelihood(self):

        uv_wavelengths = np.array([[1.0, 0.0], [1.0, 1.0], [2.0, 2.0]])

        transformer = al.Transformer(
            uv_wavelengths=uv_wavelengths,
            grid_radians=np.array([[0.0, -1.0], [0.0, 1.0], [1.0, 1.0]]),
        )

        # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

        # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

        uv_plane_data = al.UVPlaneData(
            shape=(1, 2),
            visibilities=5.0 * np.ones((3, 2)),
            pixel_scale=1.0,
            noise_map=2.0 * np.ones((3,)),
            uv_wavelengths=uv_wavelengths,
            primary_beam=None,
        )

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True, True],
                    [True, False, False, False, True],
                    [True, True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_uv_plane_data_7 = al.LensUVPlaneData(
            uv_plane_data=uv_plane_data, mask=mask
        )

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=0.001),
        )

        profile_image = g0.profile_image_from_grid(
            grid=lens_uv_plane_data_7.grid, return_in_2d=False, return_binned=True
        )
        model_visibilities_manual = transformer.visibilities_from_image_1d(
            image_1d=profile_image
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        fit = al.LensUVPlaneFit.from_lens_data_and_tracer(
            lens_data=lens_uv_plane_data_7, tracer=tracer
        )

        assert (fit._mask == np.array([False, False])).all()

        assert (fit._data == np.array([[5.0, 5.0], [5.0, 5.0], [5.0, 5.0]])).all()
        assert (
            fit.visibilities() == np.array([[[5.0, 5.0], [5.0, 5.0], [5.0, 5.0]]])
        ).all()
        assert (
            fit._noise_map == np.array([[[2.0, 2.0], [2.0, 2.0], [2.0, 2.0]]])
        ).all()
        assert (
            fit.noise_map() == np.array([[[2.0, 2.0], [2.0, 2.0], [2.0, 2.0]]])
        ).all()

        assert fit._model_data == pytest.approx(model_visibilities_manual, 1.0e-4)
        assert fit.model_visibilities() == pytest.approx(
            model_visibilities_manual, 1.0e-4
        )

        # moddel visibilities are all [1.94805, 0.0]

        assert fit._residual_map == pytest.approx(
            np.array([[3.0519, 5.0], [3.0519, 5.0], [3.0519, 5.0]]), 1.0e-4
        )
        assert fit.residual_map() == pytest.approx(
            np.array([[3.0519, 5.0], [3.0519, 5.0], [3.0519, 5.0]]), 1.0e-4
        )
        assert fit._normalized_residual_map == pytest.approx(
            np.array(
                [
                    [3.0519 / 2.0, 5.0 / 2.0],
                    [3.0519 / 2.0, 5.0 / 2.0],
                    [3.0519 / 2.0, 5.0 / 2.0],
                ]
            ),
            1.0e-4,
        )
        assert fit.normalized_residual_map() == pytest.approx(
            np.array(
                [
                    [3.0519 / 2.0, 5.0 / 2.0],
                    [3.0519 / 2.0, 5.0 / 2.0],
                    [3.0519 / 2.0, 5.0 / 2.0],
                ]
            ),
            1.0e-4,
        )
        assert fit._chi_squared_map == pytest.approx(
            np.array(
                [
                    [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                    [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                    [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                ]
            ),
            1.0e-4,
        )
        assert fit.chi_squared_map() == pytest.approx(
            np.array(
                [
                    [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                    [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                    [(3.0519 / 2.0) ** 2.0, (5.0 / 2.0) ** 2.0],
                ]
            ),
            1.0e-4,
        )

        assert fit.chi_squared == pytest.approx(25.73579, 1.0e-4)
        assert fit.reduced_chi_squared == pytest.approx(25.73579 / 6.0)
        assert fit.noise_normalization == pytest.approx(
            (6.0 * np.log(2 * np.pi * 2.0 ** 2.0)), 1.0e-4
        )
        assert fit.likelihood == pytest.approx(
            -0.5 * (25.73579 + 6.0 * np.log(2 * np.pi * 2.0 ** 2.0)), 1.0e-4
        )


class TestCompareToManualProfilesOnly:
    def test___all_lens_fit_quantities__no_hyper_methods(self, lens_uv_plane_data_7):

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
            mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
        )

        g1 = al.Galaxy(
            redshift=1.0,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

        fit = al.LensUVPlaneFit.from_lens_data_and_tracer(
            lens_data=lens_uv_plane_data_7, tracer=tracer
        )

        assert lens_uv_plane_data_7.noise_map(return_x2=True) == pytest.approx(
            fit.noise_map()
        )

        model_visibilities = tracer.profile_visibilities_from_grid_and_transformer(
            grid=lens_uv_plane_data_7.grid, transformer=lens_uv_plane_data_7.transformer
        )

        assert model_visibilities == pytest.approx(fit._model_data, 1e-4)
        assert model_visibilities == pytest.approx(fit.model_visibilities(), 1e-4)

        residual_map = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=lens_uv_plane_data_7.visibilities(),
            mask=lens_uv_plane_data_7.visibilities_mask,
            model_data=model_visibilities,
        )

        assert residual_map == pytest.approx(fit._residual_map, 1e-4)
        assert residual_map == pytest.approx(fit.residual_map(), 1e-4)

        normalized_residual_map = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map,
            mask=lens_uv_plane_data_7.visibilities_mask,
            noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
        )

        assert normalized_residual_map == pytest.approx(
            fit._normalized_residual_map, 1e-4
        )
        assert normalized_residual_map == pytest.approx(
            fit.normalized_residual_map(), 1e-4
        )

        chi_squared_map = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map,
            mask=lens_uv_plane_data_7.visibilities_mask,
            noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
        )

        assert chi_squared_map == pytest.approx(fit._chi_squared_map, 1e-4)
        assert chi_squared_map == pytest.approx(fit.chi_squared_map(), 1e-4)

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=fit._chi_squared_map,
            mask=lens_uv_plane_data_7.visibilities_mask,
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
            mask=lens_uv_plane_data_7.visibilities_mask,
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)
        assert likelihood == fit.figure_of_merit

    # def test___lens_fit_galaxy_image_dict__corresponds_to_galaxy_images(
    #     self, lens_uv_plane_data_7
    # ):
    #     g0 = al.Galaxy(
    #         redshift=0.5,
    #         light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
    #         mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
    #     )
    #     g1 = al.Galaxy(
    #         redshift=1.0,
    #         light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
    #     )
    #     g2 = al.Galaxy(redshift=1.0)
    #
    #     tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])
    #
    #     fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
    #         lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
    #     )
    #
    #     traced_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
    #         grid=lens_uv_plane_data_7.grid, return_in_2d=False
    #     )
    #     traced_blurring_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
    #         grid=lens_uv_plane_data_7.blurring_grid, return_in_2d=False
    #     )
    #
    #     g0_profile_image_1d = g0.profile_image_from_grid(
    #         grid=traced_grids_of_planes[0], return_in_2d=False, return_binned=True
    #     )
    #     g0_blurring_image_1d = g0.profile_image_from_grid(
    #         grid=traced_blurring_grids_of_planes[0],
    #         return_in_2d=False,
    #         return_binned=True,
    #     )
    #
    #     g0_blurred_profile_image_1d = lens_uv_plane_data_7.convolver.convolve_image(
    #         image_array=g0_profile_image_1d,
    #         blurring_array=g0_blurring_image_1d,
    #     )
    #
    #     g1_profile_image_1d = g1.profile_image_from_grid(
    #         grid=traced_grids_of_planes[1], return_in_2d=False, return_binned=True
    #     )
    #     g1_blurring_image_1d = g1.profile_image_from_grid(
    #         grid=traced_blurring_grids_of_planes[1],
    #         return_in_2d=False,
    #         return_binned=True,
    #     )
    #
    #     g1_blurred_profile_image_1d = lens_uv_plane_data_7.convolver.convolve_image(
    #         image_array=g1_profile_image_1d,
    #         blurring_array=g1_blurring_image_1d,
    #     )
    #
    #     assert fit.galaxy_model_visibilities_dict[g0] == pytest.approx(
    #         g0_blurred_profile_image_1d, 1.0e-4
    #     )
    #     assert fit.galaxy_model_visibilities_dict[g1] == pytest.approx(
    #         g1_blurred_profile_image_1d, 1.0e-4
    #     )
    #     assert (fit.galaxy_model_visibilities_dict[g2] == np.zeros(9)).all()
    #
    #     assert fit._model_data == pytest.approx(
    #         fit.galaxy_model_visibilities_dict[g0] + fit.galaxy_model_visibilities_dict[g1], 1.0e-4
    #     )
    #
    #     g0_blurred_profile_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
    #         array_1d=g0_blurred_profile_image_1d
    #     )
    #     g1_blurred_profile_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
    #         array_1d=g1_blurred_profile_image_1d
    #     )
    #
    #     assert fit.galaxy_model_visibilities_2d_dict[g0] == pytest.approx(
    #         g0_blurred_profile_image_2d, 1.0e-4
    #     )
    #     assert fit.galaxy_model_visibilities_2d_dict[g1] == pytest.approx(
    #         g1_blurred_profile_image_2d, 1.0e-4
    #     )
    #
    #     assert fit.model_visibilities(return_x2=True) == pytest.approx(
    #         fit.galaxy_model_visibilities_2d_dict[g0] + fit.galaxy_model_visibilities_2d_dict[g1], 1.0e-4
    #     )
    #
    # def test___model_image_and_model_visibilitiess_of_planes_and_unmasked_blurred_profile_image_properties(
    #     self, lens_uv_plane_data_7
    # ):
    #
    #     g0 = al.Galaxy(
    #         redshift=0.5,
    #         light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
    #         mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
    #     )
    #
    #     g1 = al.Galaxy(
    #         redshift=1.0,
    #         light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
    #     )
    #
    #     tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])
    #
    #     fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
    #         lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
    #     )
    #
    #     blurred_profile_image_2d_of_planes = tracer.blurred_profile_images_of_planes_from_grid_and_convolver(
    #         grid=lens_uv_plane_data_7.grid,
    #         convolver=lens_uv_plane_data_7.convolver,
    #         return_x2=True,
    #     )
    #
    #     model_visibilitiess_of_planes = fit.model_visibilitiess_of_planes(return_x2=True)
    #
    #     assert blurred_profile_image_2d_of_planes[0] == pytest.approx(
    #         model_visibilitiess_of_planes[0], 1.0e-4
    #     )
    #     assert blurred_profile_image_2d_of_planes[1] == pytest.approx(
    #         model_visibilitiess_of_planes[1], 1.0e-4
    #     )
    #
    #     unmasked_blurred_profile_image = tracer.unmasked_blurred_profile_image_from_grid_and_psf(
    #         grid=lens_uv_plane_data_7.grid, psf=lens_uv_plane_data_7.psf
    #     )
    #
    #     assert (
    #         unmasked_blurred_profile_image == fit.unmasked_blurred_profile_image
    #     ).all()
    #
    #     unmasked_blurred_profile_image_of_planes = tracer.unmasked_blurred_profile_image_of_planes_from_grid_and_psf(
    #         grid=lens_uv_plane_data_7.grid, psf=lens_uv_plane_data_7.psf
    #     )
    #
    #     assert (
    #         unmasked_blurred_profile_image_of_planes[0]
    #         == fit.unmasked_blurred_profile_image_of_planes[0]
    #     ).all()
    #     assert (
    #         unmasked_blurred_profile_image_of_planes[1]
    #         == fit.unmasked_blurred_profile_image_of_planes[1]
    #     ).all()
    #
    #     unmasked_blurred_profile_image_of_galaxies = tracer.unmasked_blurred_profile_image_of_planes_and_galaxies_from_grid_and_psf(
    #         grid=lens_uv_plane_data_7.grid, psf=lens_uv_plane_data_7.psf
    #     )
    #
    #     assert (
    #         unmasked_blurred_profile_image_of_galaxies[0][0]
    #         == fit.unmasked_blurred_profile_image_of_planes_and_galaxies[0][0]
    #     ).all()
    #     assert (
    #         unmasked_blurred_profile_image_of_galaxies[1][0]
    #         == fit.unmasked_blurred_profile_image_of_planes_and_galaxies[1][0]
    #     ).all()


#
# class TestCompareToManualInversionOnly:
#     def test___all_lens_fit_quantities__no_hyper_methods(self, lens_uv_plane_data_7):
#
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#
#         g0 = al.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)
#
#         tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
#         )
#
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, pixelization_grid=None
#         )
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             mapper=mapper,
#             regularization=reg,
#             image_1d=lens_uv_plane_data_7.visibilities(),
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#             convolver=lens_uv_plane_data_7.convolver,
#         )
#
#         assert inversion.reconstructed_data_1d == pytest.approx(
#             fit._model_data, 1e-4
#         )
#         assert inversion.reconstructed_data_2d == fit.model_visibilities(return_x2=True)
#
#         residual_map = af.fit_util.residual_map_from_data_mask_and_model_data(
#             data=lens_uv_plane_data_7.visibilities(),
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             model_data=inversion.reconstructed_data_1d,
#         )
#         residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=residual_map
#         )
#
#         assert residual_map == pytest.approx(fit._residual_map, 1e-4)
#         assert residual_map_2d == pytest.approx(fit.residual_map(return_x2=True))
#
#         normalized_residual_map = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#         )
#
#         normalized_residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=normalized_residual_map
#         )
#
#         assert normalized_residual_map == pytest.approx(
#             fit._normalized_residual_map, 1e-4
#         )
#         assert normalized_residual_map_2d == pytest.approx(
#             fit.normalized_residual_map(return_x2=True)
#         )
#
#         chi_squared_map = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#         )
#         chi_squared_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=chi_squared_map
#         )
#
#         assert chi_squared_map == pytest.approx(fit._chi_squared_map, 1e-4)
#         assert chi_squared_map_2d == pytest.approx(
#             fit.chi_squared_map(return_x2=True)
#         )
#
#         chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
#             chi_squared_map=chi_squared_map_2d, mask=lens_uv_plane_data_7.visibilities_mask
#         )
#
#         noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#         )
#
#         likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
#             chi_squared=chi_squared, noise_normalization=noise_normalization
#         )
#
#         assert likelihood == pytest.approx(fit.likelihood, 1e-4)
#
#         likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert likelihood_with_regularization == pytest.approx(
#             fit.likelihood_with_regularization, 1e-4
#         )
#
#         evidence = evidence_from_inversion_terms(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
#             log_regularization_term=inversion.log_det_regularization_matrix_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert evidence == fit.evidence
#         assert evidence == fit.figure_of_merit
#
#     def test___lens_fit_galaxy_image_dict__has_inversion_reconstructed_data(
#         self, lens_uv_plane_data_7
#     ):
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#
#         g0 = al.Galaxy(redshift=0.5)
#         g1 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)
#
#         tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
#         )
#
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, pixelization_grid=None
#         )
#
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             mapper=mapper,
#             regularization=reg,
#             image_1d=lens_uv_plane_data_7.visibilities(),
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#             convolver=lens_uv_plane_data_7.convolver,
#         )
#
#         assert (fit.galaxy_image_1d_dict[g0] == np.zeros(9)).all()
#
#         assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
#             inversion.reconstructed_data_1d, 1.0e-4
#         )
#
#         assert fit._model_data == pytest.approx(
#             fit.galaxy_image_1d_dict[g1], 1.0e-4
#         )
#
#         assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
#             inversion.reconstructed_data_2d, 1.0e-4
#         )
#
#         assert fit.model_visibilities(return_x2=True) == pytest.approx(
#             fit.galaxy_image_2d_dict[g1], 1.0e-4
#         )
#
#     def test___all_lens_fit_quantities__include_hyper_methods(self, lens_uv_plane_data_7):
#
#         hyper_noise_map_max = 0.2
#         lens_uv_plane_data_7.hyper_noise_map_max = hyper_noise_map_max
#
#         hyper_image_sky = al.HyperImageSky(sky_scale=1.0)
#
#         hyper_background_noise = al.HyperBackgroundNoise(noise_scale=1.0)
#
#         image_1d = hyper_image_sky.image_scaled_sky_from_image(
#             image=lens_uv_plane_data_7.visibilities()
#         )
#
#         hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True)
#         )
#
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#
#         g0 = al.Galaxy(
#             redshift=0.5,
#             pixelization=pix,
#             regularization=reg,
#             hyper_galaxy=al.HyperGalaxy(
#                 contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
#             ),
#             hyper_model_visibilities=np.ones(9),
#             hyper_galaxy_image_1d=np.ones(9),
#             hyper_minimum_value=0.0,
#         )
#
#         tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7,
#             tracer=tracer,
#             hyper_image_sky=hyper_image_sky,
#             hyper_background_noise=hyper_background_noise,
#         )
#
#         hyper_noise_1d = tracer.hyper_noise_map_from_noise_map(
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True)
#         )
#         hyper_noise_map = hyper_noise_map_background_1d + hyper_noise_1d
#         hyper_noise_map[
#             hyper_noise_map > hyper_noise_map_max
#         ] = hyper_noise_map_max
#
#         hyper_noise_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=hyper_noise_map
#         )
#         assert hyper_noise_map == pytest.approx(fit._noise_map, 1e-4)
#         assert hyper_noise_map_2d == pytest.approx(fit.noise_map(return_x2=True, return_masked=True))
#
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, inversion_uses_border=False
#         )
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             mapper=mapper,
#             regularization=reg,
#             image_1d=image_1d,
#             noise_map=hyper_noise_map,
#             convolver=lens_uv_plane_data_7.convolver,
#         )
#
#         assert inversion.reconstructed_data_1d == pytest.approx(
#             fit._model_data, 1e-4
#         )
#         assert inversion.reconstructed_data_2d == fit.model_visibilities(return_x2=True)
#
#         residual_map = af.fit_util.residual_map_from_data_mask_and_model_data(
#             data=image_1d,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             model_data=inversion.reconstructed_data_1d,
#         )
#         residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=residual_map
#         )
#
#         assert residual_map == pytest.approx(fit._residual_map, 1e-4)
#         assert residual_map_2d == pytest.approx(fit.residual_map(return_x2=True))
#
#         normalized_residual_map = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=hyper_noise_map,
#         )
#
#         normalized_residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=normalized_residual_map
#         )
#
#         assert normalized_residual_map == pytest.approx(
#             fit._normalized_residual_map, 1e-4
#         )
#         assert normalized_residual_map_2d == pytest.approx(
#             fit.normalized_residual_map(return_x2=True)
#         )
#
#         chi_squared_map = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=hyper_noise_map,
#         )
#         chi_squared_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=chi_squared_map
#         )
#
#         assert chi_squared_map == pytest.approx(fit._chi_squared_map, 1e-4)
#         assert chi_squared_map_2d == pytest.approx(
#             fit.chi_squared_map(return_x2=True)
#         )
#
#         chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
#             chi_squared_map=chi_squared_map_2d, mask=lens_uv_plane_data_7.visibilities_mask
#         )
#
#         noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
#             mask=lens_uv_plane_data_7.visibilities_mask, noise_map=hyper_noise_map_2d
#         )
#
#         likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
#             chi_squared=chi_squared, noise_normalization=noise_normalization
#         )
#
#         assert likelihood == pytest.approx(fit.likelihood, 1e-4)
#
#         likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert likelihood_with_regularization == pytest.approx(
#             fit.likelihood_with_regularization, 1e-4
#         )
#
#         evidence = evidence_from_inversion_terms(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
#             log_regularization_term=inversion.log_det_regularization_matrix_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert evidence == fit.evidence
#         assert evidence == fit.figure_of_merit
#
#     def test___blurred_and_model_visibilitiess_of_planes_and_unmasked_blurred_profile_image_properties(
#         self, lens_uv_plane_data_7
#     ):
#
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#
#         g0 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)
#
#         tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
#         )
#
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, inversion_uses_border=False
#         )
#
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             mapper=mapper,
#             regularization=reg,
#             image_1d=lens_uv_plane_data_7.visibilities(),
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#             convolver=lens_uv_plane_data_7.convolver,
#         )
#
#         model_visibilitiess_of_planes = fit.model_visibilitiess_of_planes(return_x2=True)
#
#         assert (model_visibilitiess_of_planes[0] == np.zeros((7, 7))).all()
#         assert inversion.reconstructed_data_2d == pytest.approx(
#             model_visibilitiess_of_planes[1], 1.0e-4
#         )
#
#
# class TestCompareToManualProfilesAndInversion:
#     def test___all_lens_fit_quantities__no_hyper_methods(self, lens_uv_plane_data_7):
#         galaxy_light = al.Galaxy(
#             redshift=0.5,
#             light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
#         )
#
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#         galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)
#
#         tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
#         )
#
#         blurred_profile_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
#             grid=lens_uv_plane_data_7.grid,
#             convolver=lens_uv_plane_data_7.convolver,
#             return_in_2d=False,
#         )
#
#         blurred_profile_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=blurred_profile_image_1d
#         )
#
#         assert blurred_profile_image_1d == pytest.approx(
#             fit.blurred_profile_image(return_in_2d=False), 1e-4
#         )
#         assert blurred_profile_image_2d == pytest.approx(
#             fit.blurred_profile_image(return_x2=True)
#         )
#
#         profile_subtracted_image_1d = (
#             lens_uv_plane_data_7.visibilities() - blurred_profile_image_1d
#         )
#         profile_subtracted_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=profile_subtracted_image_1d
#         )
#
#         assert profile_subtracted_image_1d == pytest.approx(
#             fit.profile_subtracted_image(return_in_2d=False)
#         )
#         assert profile_subtracted_image_2d == pytest.approx(
#             fit.profile_subtracted_image(return_x2=True)
#         )
#
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, inversion_uses_border=False
#         )
#
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             image_1d=profile_subtracted_image_1d,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#             convolver=lens_uv_plane_data_7.convolver,
#             mapper=mapper,
#             regularization=reg,
#         )
#
#         model_visibilities = blurred_profile_image_1d + inversion.reconstructed_data_1d
#         model_visibilities_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=model_visibilities
#         )
#
#         assert model_visibilities == pytest.approx(fit._model_data, 1e-4)
#         assert model_visibilities_2d == pytest.approx(fit.model_visibilities(return_x2=True))
#
#         residual_map = af.fit_util.residual_map_from_data_mask_and_model_data(
#             data=lens_uv_plane_data_7.visibilities(),
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             model_data=model_visibilities,
#         )
#         residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=residual_map
#         )
#
#         assert residual_map == pytest.approx(fit._residual_map, 1e-4)
#         assert residual_map_2d == pytest.approx(fit.residual_map(return_x2=True))
#
#         normalized_residual_map = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#         )
#
#         normalized_residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=normalized_residual_map
#         )
#
#         assert normalized_residual_map == pytest.approx(
#             fit._normalized_residual_map, 1e-4
#         )
#         assert normalized_residual_map_2d == pytest.approx(
#             fit.normalized_residual_map(return_x2=True)
#         )
#
#         chi_squared_map = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#         )
#         chi_squared_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=chi_squared_map
#         )
#
#         assert chi_squared_map == pytest.approx(fit._chi_squared_map, 1e-4)
#         assert chi_squared_map_2d == pytest.approx(
#             fit.chi_squared_map(return_x2=True)
#         )
#
#         chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
#             chi_squared_map=chi_squared_map_2d, mask=lens_uv_plane_data_7.visibilities_mask
#         )
#
#         noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#         )
#
#         likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
#             chi_squared=chi_squared, noise_normalization=noise_normalization
#         )
#
#         assert likelihood == pytest.approx(fit.likelihood, 1e-4)
#
#         likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert likelihood_with_regularization == pytest.approx(
#             fit.likelihood_with_regularization, 1e-4
#         )
#
#         evidence = evidence_from_inversion_terms(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
#             log_regularization_term=inversion.log_det_regularization_matrix_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert evidence == fit.evidence
#         assert evidence == fit.figure_of_merit
#
#     def test___lens_fit_galaxy_image_dict__has_blurred_profile_images_and_inversion_reconstructed_data(
#         self, lens_uv_plane_data_7
#     ):
#
#         g0 = al.Galaxy(
#             redshift=0.5,
#             light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
#         )
#         g1 = al.Galaxy(
#             redshift=0.5,
#             light_profile=al.light_profiles.EllipticalSersic(intensity=2.0),
#         )
#         g2 = al.Galaxy(redshift=0.5)
#
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#         galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)
#
#         tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2, galaxy_pix])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
#         )
#
#         traced_grids = tracer.traced_grids_of_planes_from_grid(
#             grid=lens_uv_plane_data_7.grid
#         )
#         traced_blurring_grids = tracer.traced_grids_of_planes_from_grid(
#             grid=lens_uv_plane_data_7.blurring_grid
#         )
#
#         g0_profile_image_1d = g0.profile_image_from_grid(
#             grid=traced_grids[0], return_in_2d=False, return_binned=True
#         )
#         g0_blurring_image_1d = g0.profile_image_from_grid(
#             grid=traced_blurring_grids[0], return_in_2d=False, return_binned=False
#         )
#
#         g0_blurred_profile_image_1d = lens_uv_plane_data_7.convolver.convolve_image(
#             image_array=g0_profile_image_1d,
#             blurring_array=g0_blurring_image_1d,
#         )
#
#         g1_profile_image_1d = g1.profile_image_from_grid(
#             grid=traced_grids[1], return_in_2d=False, return_binned=True
#         )
#         g1_blurring_image_1d = g1.profile_image_from_grid(
#             grid=traced_blurring_grids[1], return_in_2d=False, return_binned=False
#         )
#
#         g1_blurred_profile_image_1d = lens_uv_plane_data_7.convolver.convolve_image(
#             image_array=g1_profile_image_1d,
#             blurring_array=g1_blurring_image_1d,
#         )
#
#         blurred_profile_image_1d = (
#             g0_blurred_profile_image_1d + g1_blurred_profile_image_1d
#         )
#
#         profile_subtracted_image_1d = (
#             lens_uv_plane_data_7.visibilities() - blurred_profile_image_1d
#         )
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, inversion_uses_border=False
#         )
#
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             image_1d=profile_subtracted_image_1d,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#             convolver=lens_uv_plane_data_7.convolver,
#             mapper=mapper,
#             regularization=reg,
#         )
#
#         assert fit.galaxy_image_1d_dict[g0] == pytest.approx(
#             g0_blurred_profile_image_1d, 1.0e-4
#         )
#         assert fit.galaxy_image_1d_dict[g1] == pytest.approx(
#             g1_blurred_profile_image_1d, 1.0e-4
#         )
#         assert (fit.galaxy_image_1d_dict[g2] == np.zeros(9)).all()
#         assert fit.galaxy_image_1d_dict[galaxy_pix] == pytest.approx(
#             inversion.reconstructed_data_1d, 1.0e-4
#         )
#
#         assert fit._model_data == pytest.approx(
#             fit.galaxy_image_1d_dict[g0]
#             + fit.galaxy_image_1d_dict[g1]
#             + fit.galaxy_image_1d_dict[galaxy_pix],
#             1.0e-4,
#         )
#
#         g0_blurred_profile_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=g0_blurred_profile_image_1d
#         )
#         g1_blurred_profile_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=g1_blurred_profile_image_1d
#         )
#
#         assert fit.galaxy_image_2d_dict[g0] == pytest.approx(
#             g0_blurred_profile_image_2d, 1.0e-4
#         )
#         assert fit.galaxy_image_2d_dict[g1] == pytest.approx(
#             g1_blurred_profile_image_2d, 1.0e-4
#         )
#         assert fit.galaxy_image_2d_dict[galaxy_pix] == pytest.approx(
#             inversion.reconstructed_data_2d, 1.0e-4
#         )
#
#         assert fit.model_visibilities(return_x2=True) == pytest.approx(
#             fit.galaxy_image_2d_dict[g0]
#             + fit.galaxy_image_2d_dict[g1]
#             + inversion.reconstructed_data_2d,
#             1.0e-4,
#         )
#
#     def test___all_lens_fit_quantities__include_hyper_methods(self, lens_uv_plane_data_7):
#
#         hyper_noise_map_max = 0.2
#         lens_uv_plane_data_7.hyper_noise_map_max = hyper_noise_map_max
#
#         hyper_image_sky = al.HyperImageSky(sky_scale=1.0)
#
#         hyper_background_noise = al.HyperBackgroundNoise(noise_scale=1.0)
#
#         image_1d = hyper_image_sky.image_scaled_sky_from_image(
#             image=lens_uv_plane_data_7.visibilities()
#         )
#
#         hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True)
#         )
#
#         galaxy_light = al.Galaxy(
#             redshift=0.5,
#             light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
#             hyper_galaxy=al.HyperGalaxy(
#                 contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
#             ),
#             hyper_model_visibilities=np.ones(9),
#             hyper_galaxy_image_1d=np.ones(9),
#             hyper_minimum_value=0.0,
#         )
#
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#         galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)
#
#         tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7,
#             tracer=tracer,
#             hyper_image_sky=hyper_image_sky,
#             hyper_background_noise=hyper_background_noise,
#         )
#
#         hyper_noise_1d = tracer.hyper_noise_map_from_noise_map(
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True)
#         )
#         hyper_noise_map = hyper_noise_map_background_1d + hyper_noise_1d
#         hyper_noise_map[
#             hyper_noise_map > hyper_noise_map_max
#         ] = hyper_noise_map_max
#
#         hyper_noise_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=hyper_noise_map
#         )
#
#         assert hyper_noise_map == pytest.approx(fit._noise_map, 1e-4)
#         assert hyper_noise_map_2d == pytest.approx(fit.noise_map(return_x2=True, return_masked=True))
#
#         blurred_profile_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
#             grid=lens_uv_plane_data_7.grid,
#             convolver=lens_uv_plane_data_7.convolver,
#             return_in_2d=False,
#         )
#
#         blurred_profile_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=blurred_profile_image_1d
#         )
#
#         assert blurred_profile_image_1d == pytest.approx(
#             fit.blurred_profile_image(return_in_2d=False), 1e-4
#         )
#         assert blurred_profile_image_2d == pytest.approx(
#             fit.blurred_profile_image(return_x2=True)
#         )
#
#         profile_subtracted_image_1d = image_1d - blurred_profile_image_1d
#         profile_subtracted_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=profile_subtracted_image_1d
#         )
#
#         assert profile_subtracted_image_1d == pytest.approx(
#             fit.profile_subtracted_image(return_in_2d=False)
#         )
#         assert profile_subtracted_image_2d == pytest.approx(
#             fit.profile_subtracted_image(return_x2=True)
#         )
#
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, inversion_uses_border=False
#         )
#
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             image_1d=profile_subtracted_image_1d,
#             noise_map=hyper_noise_map,
#             convolver=lens_uv_plane_data_7.convolver,
#             mapper=mapper,
#             regularization=reg,
#         )
#
#         model_visibilities = blurred_profile_image_1d + inversion.reconstructed_data_1d
#         model_visibilities_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=model_visibilities
#         )
#
#         assert model_visibilities == pytest.approx(fit._model_data, 1e-4)
#         assert model_visibilities_2d == pytest.approx(fit.model_visibilities(return_x2=True))
#
#         residual_map = af.fit_util.residual_map_from_data_mask_and_model_data(
#             data=image_1d, mask=lens_uv_plane_data_7.visibilities_mask, model_data=model_visibilities
#         )
#
#         residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=residual_map
#         )
#
#         assert residual_map == pytest.approx(fit._residual_map, 1e-4)
#         assert residual_map_2d == pytest.approx(fit.residual_map(return_x2=True))
#
#         normalized_residual_map = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=hyper_noise_map,
#         )
#
#         normalized_residual_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=normalized_residual_map
#         )
#
#         assert normalized_residual_map == pytest.approx(
#             fit._normalized_residual_map, 1e-4
#         )
#         assert normalized_residual_map_2d == pytest.approx(
#             fit.normalized_residual_map(return_x2=True)
#         )
#
#         chi_squared_map = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
#             residual_map=residual_map,
#             mask=lens_uv_plane_data_7.visibilities_mask,
#             noise_map=hyper_noise_map,
#         )
#
#         chi_squared_map_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=chi_squared_map
#         )
#
#         assert chi_squared_map == pytest.approx(fit._chi_squared_map, 1e-4)
#         assert chi_squared_map_2d == pytest.approx(
#             fit.chi_squared_map(return_x2=True)
#         )
#
#         chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
#             chi_squared_map=chi_squared_map_2d, mask=lens_uv_plane_data_7.visibilities_mask
#         )
#
#         noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
#             mask=lens_uv_plane_data_7.visibilities_mask, noise_map=hyper_noise_map_2d
#         )
#
#         likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
#             chi_squared=chi_squared, noise_normalization=noise_normalization
#         )
#
#         assert likelihood == pytest.approx(fit.likelihood, 1e-4)
#
#         likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert likelihood_with_regularization == pytest.approx(
#             fit.likelihood_with_regularization, 1e-4
#         )
#
#         evidence = evidence_from_inversion_terms(
#             chi_squared=chi_squared,
#             regularization_term=inversion.regularization_term,
#             log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
#             log_regularization_term=inversion.log_det_regularization_matrix_term,
#             noise_normalization=noise_normalization,
#         )
#
#         assert evidence == fit.evidence
#         assert evidence == fit.figure_of_merit
#
#     def test___blurred_and_model_visibilitiess_of_planes_and_unmasked_blurred_profile_image_properties(
#         self, lens_uv_plane_data_7
#     ):
#         galaxy_light = al.Galaxy(
#             redshift=0.5,
#             light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
#         )
#
#         pix = al.pixelizations.Rectangular(shape=(3, 3))
#         reg = al.regularization.Constant(coefficient=1.0)
#         galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)
#
#         tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])
#
#         fit = al.LensUVPlaneFit.from_lens_uv_plane_data_and_tracer(
#             lens_uv_plane_data=lens_uv_plane_data_7, tracer=tracer
#         )
#
#         blurred_profile_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
#             grid=lens_uv_plane_data_7.grid,
#             convolver=lens_uv_plane_data_7.convolver,
#             return_in_2d=False,
#         )
#
#         blurred_profile_image_2d = lens_uv_plane_data_7.mapping.scaled_array_2d_from_array_1d(
#             array_1d=blurred_profile_image_1d
#         )
#
#         profile_subtracted_image_1d = (
#             lens_uv_plane_data_7.visibilities() - blurred_profile_image_1d
#         )
#
#         mapper = pix.mapper_from_grid_and_pixelization_grid(
#             grid=lens_uv_plane_data_7.grid, inversion_uses_border=False
#         )
#
#         inversion = al.Inversion.from_data_1d_mapper_and_regularization(
#             image_1d=profile_subtracted_image_1d,
#             noise_map=lens_uv_plane_data_7.noise_map(return_x2=True),
#             convolver=lens_uv_plane_data_7.convolver,
#             mapper=mapper,
#             regularization=reg,
#         )
#
#         model_visibilitiess_of_planes = fit.model_visibilitiess_of_planes(return_x2=True)
#
#         assert blurred_profile_image_2d == pytest.approx(
#             model_visibilitiess_of_planes[0], 1.0e-4
#         )
#         assert inversion.reconstructed_data_2d == pytest.approx(
#             model_visibilitiess_of_planes[1], 1.0e-4
#        )



from astropy import cosmology as cosmo

import autofit as af
from autolens import exc
from autolens.array import mask as msk
from autolens.lens import lens_fit
from autolens.model.inversion import pixelizations as pix
from autolens.pipeline.phase import phase_extensions
from autolens.pipeline.phase.phase import AbstractPhase


def default_mask_function(image):
    return msk.Mask.circular(
        shape=image.shape, pixel_scale=image.pixel_scale, sub_size=1, radius_arcsec=3.0
    )


def isinstance_or_prior(obj, cls):
    if isinstance(obj, cls):
        return True
    if isinstance(obj, af.PriorModel) and obj.cls == cls:
        return True
    return False


class PhaseData(AbstractPhase):
    galaxies = af.PhaseProperty("galaxies")

    def __init__(
        self,
        phase_name,
        phase_tag,
        phase_folders=tuple(),
        galaxies=None,
        optimizer_class=af.MultiNest,
        cosmology=cosmo.Planck15,
        sub_size=2,
        signal_to_noise_limit=None,
        positions_threshold=None,
        mask_function=None,
        inner_mask_radii=None,
        pixel_scale_interpolation_grid=None,
        pixel_scale_binned_cluster_grid=None,
        inversion_uses_border=True,
        inversion_pixel_limit=None,
        auto_link_priors=False,
    ):

        """

        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_size: int
            The side length of the subgrid
        pixel_scale_binned_cluster_grid : float or None
            If *True*, the hyper_galaxies image used to generate the cluster'grids weight map will be binned
            up to this higher pixel scale to speed up the KMeans clustering algorithm. \
        """

        super(PhaseData, self).__init__(
            phase_name=phase_name,
            phase_tag=phase_tag,
            phase_folders=phase_folders,
            optimizer_class=optimizer_class,
            cosmology=cosmology,
            auto_link_priors=auto_link_priors,
        )

        self.sub_size = sub_size
        self.signal_to_noise_limit = signal_to_noise_limit
        self.positions_threshold = positions_threshold
        self.mask_function = mask_function
        self.inner_mask_radii = inner_mask_radii
        self.pixel_scale_interpolation_grid = pixel_scale_interpolation_grid
        self.pixel_scale_binned_cluster_grid = pixel_scale_binned_cluster_grid
        self.inversion_uses_border = inversion_uses_border

        self.inversion_pixel_limit = (
            inversion_pixel_limit
            or af.conf.instance.general.get(
                "inversion", "inversion_pixel_limit_overall", int
            )
        )
        self.hyper_noise_map_max = af.conf.instance.general.get(
            "hyper", "hyper_noise_map_max", float
        )

        self.galaxies = galaxies or []

        self.is_hyper_phase = False

    @property
    def pixelization(self):
        for galaxy in self.galaxies:
            if galaxy.pixelization is not None:
                if isinstance(galaxy.pixelization, af.PriorModel):
                    return galaxy.pixelization.cls
                else:
                    return galaxy.pixelization

    @property
    def uses_cluster_inversion(self):
        if self.galaxies:
            for galaxy in self.galaxies:
                if isinstance_or_prior(galaxy.pixelization, pix.VoronoiBrightnessImage):
                    return True
        return False

    def run(self, data, results=None, mask=None, positions=None):
        """
        Run this phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        data: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """
        analysis = self.make_analysis(
            data=data, results=results, mask=mask, positions=positions
        )

        self.variable = self.variable.populate(results)
        self.customize_priors(results)
        self.assert_and_save_pickle()

        result = self.run_analysis(analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, data, results=None, mask=None, positions=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        data: im.Imaging
            An lens_data that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """
        raise NotImplementedError()

    def setup_phase_mask(self, data, mask):

        if self.mask_function is not None:
            mask = self.mask_function(image=data.image, sub_size=self.sub_size)
        elif mask is None and self.mask_function is None:
            mask = default_mask_function(image=data.image)

        if mask.sub_size != self.sub_size:
            mask = mask.new_mask_with_new_sub_size(sub_size=self.sub_size)

        if self.inner_mask_radii is not None:
            inner_mask = msk.Mask.circular(
                shape=mask.shape,
                pixel_scale=mask.pixel_scale,
                radius_arcsec=self.inner_mask_radii,
                sub_size=self.sub_size,
                invert=True,
            )
            mask = mask + inner_mask

        return mask

    def check_positions(self, positions):

        if self.positions_threshold is not None and positions is None:
            raise exc.PhaseException(
                "You have specified for a phase to use positions, but not input positions to the "
                "pipeline when you ran it."
            )

    def pixel_scale_binned_grid_from_mask(self, mask):

        pixel_scale_binned_grid = None

        if self.uses_cluster_inversion:

            if self.pixel_scale_binned_cluster_grid is None:

                pixel_scale_binned_cluster_grid = mask.pixel_scale

            else:

                pixel_scale_binned_cluster_grid = self.pixel_scale_binned_cluster_grid

            if pixel_scale_binned_cluster_grid > mask.pixel_scale:

                bin_up_factor = int(
                    self.pixel_scale_binned_cluster_grid / mask.pixel_scale
                )

            else:

                bin_up_factor = 1

            binned_mask = mask.binned_up_mask_from_mask(bin_up_factor=bin_up_factor)

            while binned_mask.pixels_in_mask < self.inversion_pixel_limit:

                if bin_up_factor == 1:
                    raise exc.DataException(
                        f"The pixelization {self.pixelization} uses a KMeans clustering algorithm which uses "
                        f"a hyper model image to adapt the pixelization. This hyper model image must have "
                        f"more pixels than inversion pixels. Current, the inversion_pixel_limit exceeds the "
                        f"data-points in the image.\n\n To rectify this image, manually set the inversion "
                        f"pixel limit in the pipeline phases or change the inversion_pixel_limit_overall "
                        f"parameter in general.ini "
                    )

                bin_up_factor -= 1
                binned_mask = mask.binned_up_mask_from_mask(bin_up_factor=bin_up_factor)

            pixel_scale_binned_grid = mask.pixel_scale * bin_up_factor

        return pixel_scale_binned_grid

    def preload_pixelization_grids_of_planes_from_results(self, results):

        if self.is_hyper_phase:
            return None

        if (
            results is not None
            and results.last is not None
            and hasattr(results.last, "hyper_combined")
            and self.pixelization is not None
        ):
            if self.pixelization.__class__ is results.last.pixelization.__class__:
                return (
                    results.last.hyper_combined.most_likely_pixelization_grids_of_planes
                )
        return None

    def extend_with_inversion_phase(self):
        return phase_extensions.InversionPhase(phase=self)

    # noinspection PyAbstractClass
    class Analysis(AbstractPhase.Analysis):
        @property
        def lens_data(self):
            raise NotImplementedError()

        def check_positions_trace_within_threshold_via_tracer(self, tracer):

            if (
                self.lens_data.positions is not None
                and self.lens_data.positions_threshold is not None
            ):

                traced_positions_of_planes = tracer.traced_positions_of_planes_from_positions(
                    positions=self.lens_data.positions
                )

                fit = lens_fit.LensPositionFit(
                    positions=traced_positions_of_planes[-1],
                    noise_map=self.lens_data.pixel_scale,
                )

                if not fit.maximum_separation_within_threshold(
                    self.lens_data.positions_threshold
                ):
                    raise exc.RayTracingException

        def check_inversion_pixels_are_below_limit_via_tracer(self, tracer):

            if self.lens_data.inversion_pixel_limit is not None:
                pixelizations = list(filter(None, tracer.pixelizations_of_planes))
                if pixelizations:
                    for pixelization in pixelizations:
                        if pixelization.pixels > self.lens_data.inversion_pixel_limit:
                            raise exc.PixelizationException

    class Result(AbstractPhase.Result):
        @property
        def most_likely_fit(self):

            hyper_image_sky = self.analysis.hyper_image_sky_for_instance(
                instance=self.constant
            )

            hyper_background_noise = self.analysis.hyper_background_noise_for_instance(
                instance=self.constant
            )

            return self.analysis.lens_imaging_fit_for_tracer(
                tracer=self.most_likely_tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

        @property
        def mask(self):
            return self.most_likely_fit.mask

        @property
        def positions(self):
            return self.most_likely_fit.positions

        @property
        def pixelization(self):
            for galaxy in self.most_likely_fit.tracer.galaxies:
                if galaxy.pixelization is not None:
                    return galaxy.pixelization

        @property
        def most_likely_pixelization_grids_of_planes(self):
            return self.most_likely_tracer.pixelization_grids_of_planes_from_grid(
                grid=self.most_likely_fit.grid
            )[-1]



import autofit as af

from autolens.array.mapping import reshape_returned_array


class GalaxyFit(af.DataFit):
    def __init__(self, galaxy_data, model_galaxies):
        """Class which fits a set of galaxy-datas to a model galaxy, using either the galaxy's image, \
        surface-density or potential.

        Parameters
        ----------
        galaxy_data : GalaxyData
            The galaxy-datas object being fitted.
        model_galaxies : galaxy.Galaxy
            The model galaxy used to fit the galaxy-datas.
        """

        self.galaxy_data = galaxy_data
        self.model_galaxies = model_galaxies
        self.mapping = galaxy_data.mapping

        _model_data_1d = galaxy_data.profile_quantity_from_galaxies(
            galaxies=model_galaxies
        )

        super(GalaxyFit, self).__init__(
            data=galaxy_data._image_1d,
            noise_map=galaxy_data._noise_map_1d,
            mask=galaxy_data._mask_1d,
            model_data=_model_data_1d,
        )

    @property
    def grid(self):
        return self.galaxy_data.grid

    @reshape_returned_array
    def image(self, return_in_2d=True, return_masked=True):
        return self._data

    @reshape_returned_array
    def noise_map(self, return_in_2d=True, return_masked=True):
        return self._noise_map

    @property
    def mask(self):
        return self.mapping.mask

    @reshape_returned_array
    def signal_to_noise_map(self, return_in_2d=True, return_masked=True):
        return self._signal_to_noise_map

    @reshape_returned_array
    def model_image(self, return_in_2d=True):
        return self._model_data

    @reshape_returned_array
    def residual_map(self, return_in_2d=True):
        return self._residual_map

    @reshape_returned_array
    def normalized_residual_map(self, return_in_2d=True):
        return self._normalized_residual_map

    @reshape_returned_array
    def chi_squared_map(self, return_in_2d=True):
        return self._chi_squared_map

    @property
    def figure_of_merit(self):
        return self.likelihood



import logging

import numpy as np

from autolens import exc
from autolens.array.util import grid_util, array_util, mask_util, binning_util
from autolens.array import mapping
from autolens.array import scaled_array

logging.basicConfig()
logger = logging.getLogger(__name__)


class Mask(scaled_array.ScaledSquarePixelArray):
    """
    A mask represented by an ndarray where True is masked.
    """

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scale, sub_size, origin=(0.0, 0.0)):
        """ A mask, which is applied to a 2D array of hyper_galaxies to extract a set of unmasked image pixels (i.e. mask entry \
        is *False* or 0) which are then fitted in an analysis.
        
        The mask retains the pixel scale of the array and has a centre and origin.
        
        Parameters
        ----------
        array: ndarray
            An array of bools representing the mask.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        origin : (float, float)
            The (y,x) arc-second origin of the mask's coordinate system.
        centre : (float, float)
            The (y,x) arc-second centre of the mask provided it is a standard geometric shape (e.g. a circle).
        """
        # noinspection PyArgumentList
        super(Mask, self).__init__(array=array, pixel_scale=pixel_scale, origin=origin)
        self.sub_size = sub_size
        self.sub_length = int(self.sub_size ** 2.0)
        self.sub_fraction = 1.0 / self.sub_length
        self.mapping = mapping.Mapping(mask=self)

    def __array_finalize__(self, obj):
        if hasattr(obj, "sub_size"):
            self.sub_size = obj.sub_size
            self.sub_length = int(obj.sub_size ** 2.0)
            self.sub_fraction = 1.0 / obj.sub_length
        if hasattr(obj, "mapping"):
            self.mapping = obj.mapping
        if hasattr(obj, "pixel_scale"):
            self.pixel_scale = obj.pixel_scale
        if hasattr(obj, "origin"):
            self.origin = obj.origin

    def __getitem__(self, coords):
        try:
            return super(Mask, self).__getitem__(coords)
        except IndexError:
            return True

    @classmethod
    def unmasked_from_shape_pixel_scale_and_sub_size(
        cls, shape, pixel_scale, sub_size, invert=False
    ):
        """Setup a mask where all pixels are unmasked.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        mask = np.full(tuple(map(lambda d: int(d), shape)), False)
        if invert:
            mask = np.invert(mask)
        return cls(array=mask, pixel_scale=pixel_scale, sub_size=sub_size)

    @classmethod
    def circular(
        cls,
        shape,
        pixel_scale,
        radius_arcsec,
        sub_size,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """Setup a mask where unmasked pixels are within a circle of an input arc second radius and centre.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        radius_arcsec : float
            The radius (in arc seconds) of the circle within which pixels unmasked.
        centre: (float, float)
            The centre of the circle used to mask pixels.
        """
        mask = mask_util.mask_circular_from_shape_pixel_scale_and_radius(
            shape=shape,
            pixel_scale=pixel_scale,
            radius_arcsec=radius_arcsec,
            centre=centre,
        )
        if invert:
            mask = np.invert(mask)
        return cls(
            array=mask.astype("bool"), pixel_scale=pixel_scale, sub_size=sub_size
        )

    @classmethod
    def circular_annular(
        cls,
        shape,
        pixel_scale,
        sub_size,
        inner_radius_arcsec,
        outer_radius_arcsec,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """Setup a mask where unmasked pixels are within an annulus of input inner and outer arc second radii and \
         centre.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius_arcsec : float
            The radius (in arc seconds) of the inner circle outside of which pixels are unmasked.
        outer_radius_arcsec : float
            The radius (in arc seconds) of the outer circle within which pixels are unmasked.
        centre: (float, float)
            The centre of the annulus used to mask pixels.
        """

        mask = mask_util.mask_circular_annular_from_shape_pixel_scale_and_radii(
            shape=shape,
            pixel_scale=pixel_scale,
            inner_radius_arcsec=inner_radius_arcsec,
            outer_radius_arcsec=outer_radius_arcsec,
            centre=centre,
        )

        if invert:
            mask = np.invert(mask)

        return cls(
            array=mask.astype("bool"), pixel_scale=pixel_scale, sub_size=sub_size
        )

    @classmethod
    def circular_anti_annular(
        cls,
        shape,
        pixel_scale,
        sub_size,
        inner_radius_arcsec,
        outer_radius_arcsec,
        outer_radius_2_arcsec,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """Setup a mask where unmasked pixels are outside an annulus of input inner and outer arc second radii, but \
        within a second outer radius, and at a given centre.

        This mask there has two distinct unmasked regions (an inner circle and outer annulus), with an inner annulus \
        of masked pixels.

        Parameters
        ----------
        shape : (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_radius_arcsec : float
            The radius (in arc seconds) of the inner circle inside of which pixels are unmasked.
        outer_radius_arcsec : float
            The radius (in arc seconds) of the outer circle within which pixels are masked and outside of which they \
            are unmasked.
        outer_radius_2_arcsec : float
            The radius (in arc seconds) of the second outer circle within which pixels are unmasked and outside of \
            which they masked.
        centre: (float, float)
            The centre of the anti-annulus used to mask pixels.
        """

        mask = mask_util.mask_circular_anti_annular_from_shape_pixel_scale_and_radii(
            shape=shape,
            pixel_scale=pixel_scale,
            inner_radius_arcsec=inner_radius_arcsec,
            outer_radius_arcsec=outer_radius_arcsec,
            outer_radius_2_arcsec=outer_radius_2_arcsec,
            centre=centre,
        )

        if invert:
            mask = np.invert(mask)

        return cls(
            array=mask.astype("bool"), pixel_scale=pixel_scale, sub_size=sub_size
        )

    @classmethod
    def elliptical(
        cls,
        shape,
        pixel_scale,
        major_axis_radius_arcsec,
        axis_ratio,
        phi,
        sub_size,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """ Setup a mask where unmasked pixels are within an ellipse of an input arc second major-axis and centre.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the ellipse within which pixels are unmasked.
        axis_ratio : float
            The axis-ratio of the ellipse within which pixels are unmasked.
        phi : float
            The rotation angle of the ellipse within which pixels are unmasked, (counter-clockwise from the positive \
             x-axis).
        centre: (float, float)
            The centre of the ellipse used to mask pixels.
        """

        mask = mask_util.mask_elliptical_from_shape_pixel_scale_and_radius(
            shape, pixel_scale, major_axis_radius_arcsec, axis_ratio, phi, centre
        )

        if invert:
            mask = np.invert(mask)

        return cls(
            array=mask.astype("bool"), pixel_scale=pixel_scale, sub_size=sub_size
        )

    @classmethod
    def elliptical_annular(
        cls,
        shape,
        pixel_scale,
        sub_size,
        inner_major_axis_radius_arcsec,
        inner_axis_ratio,
        inner_phi,
        outer_major_axis_radius_arcsec,
        outer_axis_ratio,
        outer_phi,
        centre=(0.0, 0.0),
        invert=False,
    ):
        """Setup a mask where unmasked pixels are within an elliptical annulus of input inner and outer arc second \
        major-axis and centre.

        Parameters
        ----------
        shape: (int, int)
            The (y,x) shape of the mask in units of pixels.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        inner_major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the inner ellipse within which pixels are masked.
        inner_axis_ratio : float
            The axis-ratio of the inner ellipse within which pixels are masked.
        inner_phi : float
            The rotation angle of the inner ellipse within which pixels are masked, (counter-clockwise from the \
            positive x-axis).
        outer_major_axis_radius_arcsec : float
            The major-axis (in arc seconds) of the outer ellipse within which pixels are unmasked.
        outer_axis_ratio : float
            The axis-ratio of the outer ellipse within which pixels are unmasked.
        outer_phi : float
            The rotation angle of the outer ellipse within which pixels are unmasked, (counter-clockwise from the \
            positive x-axis).
        centre: (float, float)
            The centre of the elliptical annuli used to mask pixels.
        """

        mask = mask_util.mask_elliptical_annular_from_shape_pixel_scale_and_radius(
            shape=shape,
            pixel_scale=pixel_scale,
            inner_major_axis_radius_arcsec=inner_major_axis_radius_arcsec,
            inner_axis_ratio=inner_axis_ratio,
            inner_phi=inner_phi,
            outer_major_axis_radius_arcsec=outer_major_axis_radius_arcsec,
            outer_axis_ratio=outer_axis_ratio,
            outer_phi=outer_phi,
            centre=centre,
        )

        if invert:
            mask = np.invert(mask)

        return cls(
            array=mask.astype("bool"), pixel_scale=pixel_scale, sub_size=sub_size
        )

    @classmethod
    def mask_from_fits(cls, file_path, hdu, pixel_scale, sub_size, origin=(0.0, 0.0)):
        """
        Loads the image from a .fits file.

        Parameters
        ----------
        file_path : str
            The full path of the fits file.
        hdu : int
            The HDU number in the fits file containing the image image.
        pixel_scale: float
            The arc-second to pixel conversion factor of each pixel.
        """
        return Mask(
            array_util.numpy_array_2d_from_fits(file_path=file_path, hdu=hdu),
            pixel_scale=pixel_scale,
            sub_size=sub_size,
            origin=origin,
        )

    def new_mask_with_new_sub_size(self, sub_size):
        return Mask(
            array=self,
            pixel_scale=self.pixel_scale,
            sub_size=sub_size,
            origin=self.origin,
        )

    @property
    def sub_mask(self):

        sub_shape = (self.shape[0] * self.sub_size, self.shape[1] * self.sub_size)

        return mask_util.mask_from_shape_and_mask_1d_index_tomask_index(
            shape=sub_shape,
            mask_1d_index_tomask_index=self.mapping.sub_mask_1d_index_to_submask_index,
        )

    def binned_up_mask_from_mask(self, bin_up_factor):

        binned_up_mask = binning_util.binned_upmask_frommask_and_bin_up_factor(
            mask_2d=self, bin_up_factor=bin_up_factor
        )

        return Mask(
            array=binned_up_mask,
            pixel_scale=self.pixel_scale * bin_up_factor,
            sub_size=self.sub_size,
            origin=self.origin,
        )

    def binned_up_mask_sub_size_1_from_mask(self, bin_up_factor):

        binned_up_mask = binning_util.binned_upmask_frommask_and_bin_up_factor(
            mask_2d=self, bin_up_factor=bin_up_factor
        )

        return Mask(
            array=binned_up_mask,
            pixel_scale=self.pixel_scale * bin_up_factor,
            sub_size=1,
            origin=self.origin,
        )

    @property
    @array_util.Memoizer()
    def centre(self):
        return grid_util.grid_centre_from_grid_1d(grid_1d=self.masked_grid_1d)

    @property
    def pixels_in_mask(self):
        return int(np.size(self) - np.sum(self))

    @array_util.Memoizer()
    def blurring_mask_from_psf_shape(self, psf_shape):
        """Compute a blurring mask, which represents all masked pixels whose light will be blurred into unmasked \
        pixels via PSF convolution (see grid.Grid.blurring_grid_from_mask_and_psf_shape).

        Parameters
        ----------
        psf_shape : (int, int)
           The shape of the psf which defines the blurring region (e.g. the shape of the PSF)
        """

        if psf_shape[0] % 2 == 0 or psf_shape[1] % 2 == 0:
            raise exc.MaskException("psf_size of exterior region must be odd")

        blurring_mask = mask_util.blurring_mask_from_mask_and_psf_shape(self, psf_shape)

        return Mask(array=blurring_mask, pixel_scale=self.pixel_scale, sub_size=1)

    @property
    def edge_1d_indexes(self):
        """The indicies of the mask's edge pixels, where an edge pixel is any unmasked pixel on its edge \
        (next to at least one pixel with a *True* value).
        """
        return mask_util.edge_1d_indexes_from_mask(mask=self).astype("int")

    @property
    def border_1d_indexes(self):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        return mask_util.border_1d_indexes_from_mask(mask=self).astype("int")

    @property
    def sub_border_1d_indexes(self):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        return mask_util.sub_border_pixel_1d_indexes_from_mask_and_sub_size(
            mask=self, sub_size=self.sub_size
        ).astype("int")

    @property
    def border_grid_1d(self):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        return self.masked_grid_1d[self.border_1d_indexes]

    @property
    def sub_border_grid_1d(self):
        """The indicies of the mask's border pixels, where a border pixel is any unmasked pixel on an
        exterior edge (e.g. next to at least one pixel with a *True* value but not central pixels like those within \
        an annulus mask).
        """
        return self.masked_sub_grid_1d[self.sub_border_1d_indexes]

    @property
    def masked_grid_1d(self):
        return grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=self, pixel_scales=self.pixel_scales, sub_size=1, origin=self.origin
        )

    @property
    def masked_sub_grid_1d(self):
        return grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=self,
            pixel_scales=self.pixel_scales,
            sub_size=self.sub_size,
            origin=self.origin,
        )

    @property
    def zoom_centre(self):
        extraction_grid_1d = self.grid_arcsec_to_grid_pixels(
            grid_arcsec=self.masked_grid_1d
        )
        y_pixels_max = np.max(extraction_grid_1d[:, 0])
        y_pixels_min = np.min(extraction_grid_1d[:, 0])
        x_pixels_max = np.max(extraction_grid_1d[:, 1])
        x_pixels_min = np.min(extraction_grid_1d[:, 1])
        return (
            ((y_pixels_max + y_pixels_min - 1.0) / 2.0),
            ((x_pixels_max + x_pixels_min - 1.0) / 2.0),
        )

    @property
    def zoom_offset_pixels(self):
        return (
            self.zoom_centre[0] - self.central_pixel_coordinates[0],
            self.zoom_centre[1] - self.central_pixel_coordinates[1],
        )

    @property
    def zoom_offset_arcsec(self):
        return (
            -self.pixel_scale * self.zoom_offset_pixels[0],
            self.pixel_scale * self.zoom_offset_pixels[1],
        )

    @property
    def zoom_region(self):
        """The zoomed rectangular region corresponding to the square encompassing all unmasked values. This zoomed
        extraction region is a squuare, even if the mask is rectangular.

        This is used to zoom in on the region of an image that is used in an analysis for visualization."""

        # Have to convert mask to bool for invert function to work.
        where = np.array(np.where(np.invert(self.astype("bool"))))
        y0, x0 = np.amin(where, axis=1)
        y1, x1 = np.amax(where, axis=1)

        ylength = y1 - y0
        xlength = x1 - x0

        if ylength > xlength:
            length_difference = ylength - xlength
            x1 += int(length_difference / 2.0)
            x0 -= int(length_difference / 2.0)
        elif xlength > ylength:
            length_difference = xlength - ylength
            y1 += int(length_difference / 2.0)
            y0 -= int(length_difference / 2.0)

        return [y0, y1 + 1, x0, x1 + 1]


def load_mask_from_fits(mask_path, pixel_scale, sub_size=1, mask_hdu=0):
    return Mask.mask_from_fits(
        file_path=mask_path, hdu=mask_hdu, pixel_scale=pixel_scale, sub_size=sub_size
    )


def output_mask_to_fits(mask, mask_path, overwrite=False):
    array_util.numpy_array_2d_to_fits(
        array_2d=mask, file_path=mask_path, overwrite=overwrite
    )



import autofit as af
from autolens.model.hyper import hyper_data as hd
from autolens.model.inversion import pixelizations as px
from autolens.model.inversion import regularization as rg
from autolens.pipeline.phase.phase import AbstractPhase
from autolens.pipeline.phase.phase_imaging import PhaseImaging
from .hyper_phase import HyperPhase


# noinspection PyAbstractClass
class VariableFixingHyperPhase(HyperPhase):
    def __init__(self, phase: AbstractPhase, hyper_name: str, variable_classes=tuple()):
        super().__init__(phase=phase, hyper_name=hyper_name)
        self.variable_classes = variable_classes

    def make_hyper_phase(self):
        phase = super().make_hyper_phase()

        phase.optimizer.const_efficiency_mode = af.conf.instance.non_linear.get(
            "MultiNest", "extension_inversion_const_efficiency_mode", bool
        )
        phase.optimizer.sampling_efficiency = af.conf.instance.non_linear.get(
            "MultiNest", "extension_inversion_sampling_efficiency", float
        )
        phase.optimizer.n_live_points = af.conf.instance.non_linear.get(
            "MultiNest", "extension_inversion_n_live_points", int
        )
        phase.optimizer.multimodal = af.conf.instance.non_linear.get(
            "MultiNest", "extension_inversion_multimodal", bool
        )

        return phase

    def make_variable(self, constant):
        return constant.as_variable(self.variable_classes)

    def run_hyper(self, data, results=None, **kwargs):
        """
        Run the phase, overriding the optimizer's variable instance with one created to
        only fit pixelization hyperparameters.
        """
        phase = self.make_hyper_phase()
        phase.optimizer.variable = self.make_variable(results.last.constant)

        return phase.run(data, results=results, mask=results.last.mask)


class InversionPhase(VariableFixingHyperPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(
        self,
        phase: AbstractPhase,
        variable_classes=(px.Pixelization, rg.Regularization),
    ):
        super().__init__(
            phase=phase, variable_classes=variable_classes, hyper_name="inversion"
        )


class InversionBackgroundSkyPhase(InversionPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(self, phase: PhaseImaging):
        super().__init__(
            phase=phase,
            variable_classes=(px.Pixelization, rg.Regularization, hd.HyperImageSky),
        )


class InversionBackgroundNoisePhase(InversionPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(self, phase: PhaseImaging):
        super().__init__(
            phase=phase,
            variable_classes=(
                px.Pixelization,
                rg.Regularization,
                hd.HyperBackgroundNoise,
            ),
        )


class InversionBackgroundBothPhase(InversionPhase):
    """
    Phase that makes everything in the variable from the previous phase equal to the
    corresponding value from the best fit except for variables associated with
    pixelization
    """

    def __init__(self, phase: PhaseImaging):
        super().__init__(
            phase=phase,
            variable_classes=(
                px.Pixelization,
                rg.Regularization,
                hd.HyperImageSky,
                hd.HyperBackgroundNoise,
            ),
        )



import logging

from scipy.stats import norm

import scipy.signal
from astropy import units
from skimage.transform import resize, rescale

import numpy as np

from autolens import exc
from autolens.array import grids
from autolens.data import abstract_data
from autolens.array import scaled_array
from autolens.array.util import array_util
from autolens.array.mapping_util import grid_mapping_util
from autolens.model.profiles.light_profiles import EllipticalGaussian

logger = logging.getLogger(__name__)


class ImagingData(abstract_data.AbstractData):
    def __init__(
        self,
        image,
        pixel_scale,
        psf,
        noise_map=None,
        background_noise_map=None,
        poisson_noise_map=None,
        exposure_time_map=None,
        background_sky_map=None,
        name=None,
        **kwargs
    ):
        """A collection of 2D Imaging data (an image, noise-map, psf, etc.)

        Parameters
        ----------
        image : scaled_array.ScaledArraySquarePixels
            The array of the image data_type, in units of electrons per second.
        pixel_scale : float
            The size of each pixel in arc seconds.
        psf : PSF
            An array describing the PSF kernel of the image.
        noise_map : NoiseMap | float | ndarray
            An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
            second.
        background_noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel due to the background sky noise_map,
            preferably in units of electrons per second.
        poisson_noise_map : NoiseMap
            An array describing the RMS standard deviation error in each pixel due to the Poisson counts of the source,
            preferably in units of electrons per second.
        exposure_time_map : scaled_array.ScaledSquarePixelArray
            An array describing the effective exposure time in each imaging pixel.
        background_sky_map : scaled_array.ScaledSquarePixelArray
            An array describing the background sky.
        """

        super(ImagingData, self).__init__(
            data=image,
            pixel_scale=pixel_scale,
            noise_map=noise_map,
            exposure_time_map=exposure_time_map,
            origin=(0.0, 0.0),
        )

        self.psf = psf
        self.name = name
        self.background_noise_map = background_noise_map
        self.poisson_noise_map = poisson_noise_map
        self.background_sky_map = background_sky_map

    @property
    def image(self):
        return self._data

    @property
    def shape(self):
        return self.image.shape

    def new_imaging_data_with_binned_up_arrays(self, bin_up_factor):

        image = self.bin_up_scaled_array(
            scaled_array=self.image, bin_up_factor=bin_up_factor, method="mean"
        )
        psf = self.psf.new_psf_with_rescaled_odd_dimensioned_array(
            rescale_factor=1.0 / bin_up_factor, renormalize=True
        )
        noise_map = self.bin_up_scaled_array(
            scaled_array=self.noise_map,
            bin_up_factor=bin_up_factor,
            method="quadrature",
        )
        background_noise_map = self.bin_up_scaled_array(
            scaled_array=self.background_noise_map,
            bin_up_factor=bin_up_factor,
            method="quadrature",
        )
        poisson_noise_map = self.bin_up_scaled_array(
            scaled_array=self.poisson_noise_map,
            bin_up_factor=bin_up_factor,
            method="quadrature",
        )
        exposure_time_map = self.bin_up_scaled_array(
            scaled_array=self.exposure_time_map,
            bin_up_factor=bin_up_factor,
            method="sum",
        )
        background_sky_map = self.bin_up_scaled_array(
            scaled_array=self.background_sky_map,
            bin_up_factor=bin_up_factor,
            method="mean",
        )

        return ImagingData(
            image=image,
            pixel_scale=self.pixel_scale * bin_up_factor,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def new_imaging_data_with_resized_arrays(
        self, new_shape, new_centre_pixels=None, new_centre_arcsec=None
    ):

        image = self.resize_scaled_array(
            scaled_array=self.image,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        noise_map = self.resize_scaled_array(
            scaled_array=self.noise_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        background_noise_map = self.resize_scaled_array(
            scaled_array=self.background_noise_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        poisson_noise_map = self.resize_scaled_array(
            scaled_array=self.poisson_noise_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        exposure_time_map = self.resize_scaled_array(
            scaled_array=self.exposure_time_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        background_sky_map = self.resize_scaled_array(
            scaled_array=self.background_sky_map,
            new_shape=new_shape,
            new_centre_pixels=new_centre_pixels,
            new_centre_arcsec=new_centre_arcsec,
        )

        return ImagingData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def new_imaging_data_with_resized_psf(self, new_shape):
        psf = self.resize_scaled_array(scaled_array=self.psf, new_shape=new_shape)
        return ImagingData(
            image=self.image,
            pixel_scale=self.pixel_scale,
            psf=psf,
            noise_map=self.noise_map,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=self.background_sky_map,
            name=self.name,
        )

    def new_imaging_data_with_modified_image(self, modified_image):

        return ImagingData(
            image=modified_image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=self.noise_map,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=self.background_sky_map,
            name=self.name,
        )

    def new_imaging_data_with_poisson_noise_added(self, seed=-1):

        image_with_sky = self.image + self.background_sky_map

        image_with_sky_and_noise = image_with_sky + generate_poisson_noise(
            image=image_with_sky, exposure_time_map=self.exposure_time_map, seed=seed
        )

        image_with_noise = image_with_sky_and_noise - self.background_sky_map

        return ImagingData(
            image=image_with_noise,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=self.noise_map,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            name=self.name,
        )

    def new_imaging_data_converted_from_electrons(self):

        image = self.array_from_counts_to_electrons_per_second(array=self.image)
        noise_map = self.array_from_counts_to_electrons_per_second(array=self.noise_map)
        background_noise_map = self.array_from_counts_to_electrons_per_second(
            array=self.background_noise_map
        )
        poisson_noise_map = self.array_from_counts_to_electrons_per_second(
            array=self.poisson_noise_map
        )
        background_sky_map = self.array_from_counts_to_electrons_per_second(
            array=self.background_sky_map
        )

        return ImagingData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def new_imaging_data_converted_from_adus(self, gain):

        image = self.array_from_adus_to_electrons_per_second(
            array=self.image, gain=gain
        )
        noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.noise_map, gain=gain
        )
        background_noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.background_noise_map, gain=gain
        )
        poisson_noise_map = self.array_from_adus_to_electrons_per_second(
            array=self.poisson_noise_map, gain=gain
        )
        background_sky_map = self.array_from_adus_to_electrons_per_second(
            array=self.background_sky_map, gain=gain
        )

        return ImagingData(
            image=image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=background_sky_map,
            name=self.name,
        )

    def new_imaging_data_with_signal_to_noise_limit(self, signal_to_noise_limit):

        noise_map_limit = np.where(
            self.signal_to_noise_map > signal_to_noise_limit,
            np.abs(self.image) / signal_to_noise_limit,
            self.noise_map,
        )

        return ImagingData(
            image=self.image,
            pixel_scale=self.pixel_scale,
            psf=self.psf,
            noise_map=noise_map_limit,
            background_noise_map=self.background_noise_map,
            poisson_noise_map=self.poisson_noise_map,
            exposure_time_map=self.exposure_time_map,
            background_sky_map=self.background_sky_map,
            name=self.name,
        )

    @property
    def background_noise_map_counts(self):
        """ The background noise_maps mappers in units of counts."""
        return self.array_from_electrons_per_second_to_counts(self.background_noise_map)

    @property
    def estimated_noise_map_counts(self):
        """ The estimated noise_maps mappers of the image (using its background noise_maps mappers and image values
        in counts) in counts.
        """
        return np.sqrt(
            (np.abs(self.image_counts) + np.square(self.background_noise_map_counts))
        )

    @property
    def estimated_noise_map(self):
        """ The estimated noise_maps mappers of the image (using its background noise_maps mappers and image values
        in counts) in electrons per second.
        """
        return self.array_from_counts_to_electrons_per_second(
            self.estimated_noise_map_counts
        )

    def background_noise_from_edges(self, no_edges):
        """Estimate the background signal_to_noise_ratio by binning data_to_image located at the edge(s) of an image
        into a histogram and fitting a Gaussian profiles to this histogram. The standard deviation (sigma) of this
        Gaussian gives a signal_to_noise_ratio estimate.

        Parameters
        ----------
        no_edges : int
            Number of edges used to estimate the background signal_to_noise_ratio.

        """

        edges = []

        for edge_no in range(no_edges):
            top_edge = self.image[edge_no, edge_no : self.image.shape[1] - edge_no]
            bottom_edge = self.image[
                self.image.shape[0] - 1 - edge_no,
                edge_no : self.image.shape[1] - edge_no,
            ]
            left_edge = self.image[
                edge_no + 1 : self.image.shape[0] - 1 - edge_no, edge_no
            ]
            right_edge = self.image[
                edge_no + 1 : self.image.shape[0] - 1 - edge_no,
                self.image.shape[1] - 1 - edge_no,
            ]

            edges = np.concatenate(
                (edges, top_edge, bottom_edge, right_edge, left_edge)
            )

        return norm.fit(edges)[1]

    def __array_finalize__(self, obj):
        if isinstance(obj, ImagingData):
            try:
                self.image = obj.image
                self.pixel_scale = obj.pixel_scale
                self.psf = obj.psf
                self.noise_map = obj.noise_map
                self.background_noise_map = obj.background_noise_map
                self.poisson_noise_map = obj.poisson_noise_map
                self.exposure_time_map = obj.exposure_time_map
                self.background_sky_map = obj.background_sky_map
                self.origin = obj.origin
            except AttributeError:
                logger.debug(
                    "Original object in Imaging.__array_finalize__ missing one or more attributes"
                )


class NoiseMap(abstract_data.AbstractNoiseMap):
    @classmethod
    def from_image_and_background_noise_map(
        cls,
        pixel_scale,
        image,
        background_noise_map,
        exposure_time_map,
        gain=None,
        convert_from_electrons=False,
        convert_from_adus=False,
    ):

        if not convert_from_electrons and not convert_from_adus:
            return NoiseMap(
                array=np.sqrt(
                    np.abs(
                        ((background_noise_map) * exposure_time_map) ** 2.0
                        + (image) * exposure_time_map
                    )
                )
                / (exposure_time_map),
                pixel_scale=pixel_scale,
            )
        elif convert_from_electrons:
            return NoiseMap(
                array=np.sqrt(np.abs(background_noise_map ** 2.0 + image)),
                pixel_scale=pixel_scale,
            )
        elif convert_from_adus:
            return NoiseMap(
                array=np.sqrt(
                    np.abs((gain * background_noise_map) ** 2.0 + gain * image)
                )
                / gain,
                pixel_scale=pixel_scale,
            )


class PSF(scaled_array.ScaledSquarePixelArray):

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scale, renormalize=False, **kwargs):
        """
        Class storing a 2D Point Spread Function (PSF), including its blurring kernel.

        Parameters
        ----------
        array : ndarray
            The 2d PSF blurring kernel.
        renormalize : bool
            Renormalize the PSF such that he sum of kernel values total 1.0?
        """

        # noinspection PyArgumentList
        super().__init__(array=array, pixel_scale=pixel_scale)
        if renormalize:
            self[:, :] = np.divide(self, np.sum(self))

    @classmethod
    def from_no_blurring_kernel(cls, pixel_scale):

        array = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

        return PSF(array=array, pixel_scale=pixel_scale, renormalize=False)

    @classmethod
    def from_gaussian(
        cls, shape, pixel_scale, sigma, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0
    ):
        """Simulate the PSF as an elliptical Gaussian profile."""
        from autolens.model.profiles.light_profiles import EllipticalGaussian

        gaussian = EllipticalGaussian(
            centre=centre, axis_ratio=axis_ratio, phi=phi, intensity=1.0, sigma=sigma
        )

        grid = grids.Grid.from_shape_pixel_scale_and_sub_size(
            shape=shape, pixel_scale=pixel_scale, sub_size=1
        )

        gaussian = gaussian.profile_image_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        return PSF(array=gaussian, pixel_scale=pixel_scale, renormalize=True)

    @classmethod
    def from_as_gaussian_via_alma_fits_header_parameters(
        cls, shape, pixel_scale, y_stddev, x_stddev, theta, centre=(0.0, 0.0)
    ):

        x_stddev = (
            x_stddev * (units.deg).to(units.arcsec) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
        )
        y_stddev = (
            y_stddev * (units.deg).to(units.arcsec) / (2.0 * np.sqrt(2.0 * np.log(2.0)))
        )

        axis_ratio = x_stddev / y_stddev

        gaussian = EllipticalGaussian(
            centre=centre,
            axis_ratio=axis_ratio,
            phi=90.0 - theta,
            intensity=1.0,
            sigma=y_stddev,
        )

        from autolens.array import grids

        grid = grids.Grid.from_shape_pixel_scale_and_sub_size(
            shape=shape, pixel_scale=pixel_scale, sub_size=1
        )

        gaussian = gaussian.profile_image_from_grid(
            grid=grid, return_in_2d=True, return_binned=True
        )

        return PSF(array=gaussian, pixel_scale=pixel_scale, renormalize=True)

    @classmethod
    def from_fits_renormalized(cls, file_path, hdu, pixel_scale):
        """Loads a PSF from fits and renormalizes it

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PSF
        hdu : int
            The HDU the PSF is stored in the .fits file.

        Returns
        -------
        psf: PSF
            A renormalized PSF instance
        """
        psf = PSF.from_fits_with_scale(file_path, hdu, pixel_scale)
        psf[:, :] = np.divide(psf, np.sum(psf))
        return psf

    @classmethod
    def from_fits_with_scale(cls, file_path, hdu, pixel_scale):
        """
        Loads the PSF from a .fits file.

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PSF
        hdu : int
            The HDU the PSF is stored in the .fits file.
        """
        return cls(
            array=array_util.numpy_array_2d_from_fits(file_path, hdu),
            pixel_scale=pixel_scale,
        )

    def new_psf_with_rescaled_odd_dimensioned_array(
        self, rescale_factor, renormalize=True
    ):

        psf_rescaled = rescale(
            self,
            rescale_factor,
            anti_aliasing=False,
            mode="constant",
            multichannel=False,
        )

        if psf_rescaled.shape[0] % 2 == 0 and psf_rescaled.shape[1] % 2 == 0:
            psf_rescaled = resize(
                psf_rescaled,
                output_shape=(psf_rescaled.shape[0] + 1, psf_rescaled.shape[1] + 1),
                anti_aliasing=False,
                mode="constant",
            )
        elif psf_rescaled.shape[0] % 2 == 0 and psf_rescaled.shape[1] % 2 != 0:
            psf_rescaled = resize(
                psf_rescaled,
                output_shape=(psf_rescaled.shape[0] + 1, psf_rescaled.shape[1]),
                anti_aliasing=False,
                mode="constant",
            )
        elif psf_rescaled.shape[0] % 2 != 0 and psf_rescaled.shape[1] % 2 == 0:
            psf_rescaled = resize(
                psf_rescaled,
                output_shape=(psf_rescaled.shape[0], psf_rescaled.shape[1] + 1),
                anti_aliasing=False,
                mode="constant",
            )

        pixel_scale_factors = (
            self.shape[0] / psf_rescaled.shape[0],
            self.shape[1] / psf_rescaled.shape[1],
        )
        pixel_scale = (
            self.pixel_scale * pixel_scale_factors[0],
            self.pixel_scale * pixel_scale_factors[1],
        )
        return PSF(
            array=psf_rescaled, pixel_scale=np.max(pixel_scale), renormalize=renormalize
        )

    def new_psf_with_renormalized_array(self):
        """Renormalize the PSF such that its data_vector values sum to unity."""
        return PSF(array=self, pixel_scale=self.pixel_scale, renormalize=True)

    def convolve(self, array_2d):
        """
        Convolve an array with this PSF

        Parameters
        ----------
        image : ndarray
            An array representing the image the PSF is convolved with.

        Returns
        -------
        convolved_image : ndarray
            An array representing the image after convolution.

        Raises
        ------
        KernelException if either PSF psf dimension is odd
        """
        if self.shape[0] % 2 == 0 or self.shape[1] % 2 == 0:
            raise exc.ConvolutionException("PSF Kernel must be odd")

        return scipy.signal.convolve2d(array_2d, self, mode="same")


class PoissonNoiseMap(NoiseMap):
    @classmethod
    def from_image_and_exposure_time_map(
        cls,
        pixel_scale,
        image,
        exposure_time_map,
        gain=None,
        convert_from_electrons=False,
        convert_from_adus=False,
    ):
        if not convert_from_electrons and not convert_from_adus:
            return PoissonNoiseMap(
                array=np.sqrt(np.abs(image) * exposure_time_map) / (exposure_time_map),
                pixel_scale=pixel_scale,
            )
        elif convert_from_electrons:
            return PoissonNoiseMap(
                array=np.sqrt(np.abs(image)), pixel_scale=pixel_scale
            )
        elif convert_from_adus:
            return NoiseMap(
                array=np.sqrt(gain * np.abs(image)) / gain, pixel_scale=pixel_scale
            )


class SimulatedImagingData(ImagingData):
    def __init__(
        self,
        image,
        pixel_scale,
        psf,
        noise_map=None,
        background_noise_map=None,
        poisson_noise_map=None,
        exposure_time_map=None,
        background_sky_map=None,
        noise_realization=None,
        name=None,
        **kwargs
    ):

        super(SimulatedImagingData, self).__init__(
            image=image,
            pixel_scale=pixel_scale,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=name,
            kwargs=kwargs,
        )

        self.noise_realization = noise_realization

    @classmethod
    def from_deflections_galaxies_and_exposure_arrays(
        cls,
        deflections,
        pixel_scale,
        galaxies,
        exposure_time,
        psf=None,
        exposure_time_map=None,
        background_sky_level=0.0,
        background_sky_map=None,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        name=None,
    ):

        shape = (deflections.shape[0], deflections.shape[1])

        grid_1d = grids.Grid.from_shape_pixel_scale_and_sub_size(
            shape=shape, pixel_scale=pixel_scale, sub_size=1
        )

        deflections_1d = grid_mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_size(
            sub_grid_2d=deflections,
            mask=np.full(shape=shape, fill_value=False),
            sub_size=1,
        )

        deflected_grid_1d = grid_1d - deflections_1d

        image_2d = sum(
            map(
                lambda g: g.profile_image_from_grid(
                    grid=deflected_grid_1d, return_in_2d=True, return_binned=False
                ),
                galaxies,
            )
        )

        return cls.from_image_and_exposure_arrays(
            image=image_2d,
            pixel_scale=pixel_scale,
            exposure_time=exposure_time,
            psf=psf,
            exposure_time_map=exposure_time_map,
            background_sky_level=background_sky_level,
            background_sky_map=background_sky_map,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
            name=name,
        )

    @classmethod
    def from_tracer_grid_and_exposure_arrays(
        cls,
        tracer,
        grid,
        pixel_scale,
        exposure_time,
        psf=None,
        exposure_time_map=None,
        background_sky_level=0.0,
        background_sky_map=None,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        name=None,
    ):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        if psf is not None:
            image_2d = tracer.padded_profile_image_2d_from_grid_and_psf_shape(
                grid=grid, psf_shape=psf.shape
            )
        else:
            image_2d = tracer.profile_image_from_grid(
                grid=grid, return_in_2d=True, return_binned=True, bypass_decorator=False
            )

        return cls.from_image_and_exposure_arrays(
            image=image_2d,
            pixel_scale=pixel_scale,
            exposure_time=exposure_time,
            psf=psf,
            exposure_time_map=exposure_time_map,
            background_sky_level=background_sky_level,
            background_sky_map=background_sky_map,
            add_noise=add_noise,
            noise_if_add_noise_false=noise_if_add_noise_false,
            noise_seed=noise_seed,
            name=name,
        )

    @classmethod
    def from_image_and_exposure_arrays(
        cls,
        image,
        pixel_scale,
        exposure_time,
        psf=None,
        exposure_time_map=None,
        background_sky_level=0.0,
        background_sky_map=None,
        add_noise=True,
        noise_if_add_noise_false=0.1,
        noise_seed=-1,
        name=None,
    ):
        """
        Create a realistic simulated image by applying effects to a plain simulated image.

        Parameters
        ----------
        name
        image : ndarray
            The image before simulating (e.g. the lens and source galaxies before optics blurring and Imaging read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        exposure_time_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated image is blurred with.
        background_sky_map : ndarray
            The value of background sky in every image pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_maps is simulated and added to the image, based on the total counts in each image
            pixel
        noise_seed: int
            A seed for random noise_maps generation
        """

        if psf is None:
            psf = PSF.from_no_blurring_kernel(pixel_scale=pixel_scale)
            image_needs_trimming = False
        else:
            image_needs_trimming = True

        if exposure_time_map is None:

            exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
                value=exposure_time, shape=image.shape, pixel_scale=pixel_scale
            )

        if background_sky_map is None:

            background_sky_map = scaled_array.ScaledSquarePixelArray.single_value(
                value=background_sky_level, shape=image.shape, pixel_scale=pixel_scale
            )

        image += background_sky_map

        image = psf.convolve(image)

        if image_needs_trimming:
            image = trim_psf_edges(image, psf)
            exposure_time_map = trim_psf_edges(exposure_time_map, psf)
            background_sky_map = trim_psf_edges(background_sky_map, psf)

        if add_noise is True:
            noise_realization = generate_poisson_noise(
                image, exposure_time_map, noise_seed
            )
            image += noise_realization
            image_counts = np.multiply(image, exposure_time_map)
            noise_map = np.divide(np.sqrt(image_counts), exposure_time_map)
            noise_map = NoiseMap(array=noise_map, pixel_scale=pixel_scale)
        else:
            noise_map = NoiseMap.single_value(
                value=noise_if_add_noise_false,
                shape=image.shape,
                pixel_scale=pixel_scale,
            )
            noise_realization = None

        if np.isnan(noise_map).any():
            raise exc.DataException(
                "The noise-map has NaN values in it. This suggests your exposure time and / or"
                "background sky levels are too low, creating signal counts at or close to 0.0."
            )

        image -= background_sky_map

        # ESTIMATE THE BACKGROUND NOISE MAP FROM THE BACKGROUND SKY MAP

        background_noise_map_counts = np.sqrt(
            np.multiply(background_sky_map, exposure_time_map)
        )
        background_noise_map = np.divide(background_noise_map_counts, exposure_time_map)

        # ESTIMATE THE POISSON NOISE MAP FROM THE IMAGE

        image_counts = np.multiply(image, exposure_time_map)
        poisson_noise_map = np.divide(np.sqrt(np.abs(image_counts)), exposure_time_map)

        image = scaled_array.ScaledSquarePixelArray(
            array=image, pixel_scale=pixel_scale
        )
        background_noise_map = NoiseMap(
            array=background_noise_map, pixel_scale=pixel_scale
        )
        poisson_noise_map = PoissonNoiseMap(
            array=poisson_noise_map, pixel_scale=pixel_scale
        )

        return SimulatedImagingData(
            image,
            pixel_scale=pixel_scale,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            noise_realization=noise_realization,
            name=name,
        )

    def __array_finalize__(self, obj):
        if isinstance(obj, SimulatedImagingData):
            try:
                self.image = obj.image
                self.pixel_scale = obj.pixel_scale
                self.psf = obj.psf
                self.noise_map = obj.noise_map
                self.background_noise_map = obj.background_noise_map
                self.poisson_noise_map = obj.poisson_noise_map
                self.exposure_time_map = obj.exposure_time_map
                self.background_sky_map = obj.background_sky_map
                self.background_noise_realization = obj.background_noise_realization
                self.poisson_noise_realization = obj.poisson_noise_realization
                self.origin = obj.origin
            except AttributeError:
                logger.debug(
                    "Original object in Imaging.__array_finalize__ missing one or more attributes"
                )


def setup_random_seed(seed):
    """Setup the random seed. If the input seed is -1, the code will use a random seed for every run. If it is \
    positive, that seed is used for all runs, thereby giving reproducible results.

    Parameters
    ----------
    seed : int
        The seed of the random number generator.
    """
    if seed == -1:
        seed = np.random.randint(
            0, int(1e9)
        )  # Use one seed, so all regions have identical column non-uniformity.
    np.random.seed(seed)


def generate_poisson_noise(image, exposure_time_map, seed=-1):
    """
    Generate a two-dimensional poisson noise_maps-mappers from an image.

    Values are computed from a Poisson distribution using the image's input values in units of counts.

    Parameters
    ----------
    image : ndarray
        The 2D image, whose values in counts are used to draw Poisson noise_maps values.
    exposure_time_map : Union(ndarray, int)
        2D array of the exposure time in each pixel used to convert to / from counts and electrons per second.
    seed : int
        The seed of the random number generator, used for the random noise_maps maps.

    Returns
    -------
    poisson_noise_map: ndarray
        An array describing simulated poisson noise_maps
    """
    setup_random_seed(seed)
    image_counts = np.multiply(image, exposure_time_map)
    return image - np.divide(
        np.random.poisson(image_counts, image.shape), exposure_time_map
    )


def trim_psf_edges(array, psf):

    if psf is not None:
        psf_cut_x = np.int(np.ceil(psf.shape[0] / 2)) - 1
        psf_cut_y = np.int(np.ceil(psf.shape[1] / 2)) - 1
        array_x = np.int(array.shape[0])
        array_y = np.int(array.shape[1])
        return array[psf_cut_x : array_x - psf_cut_x, psf_cut_y : array_y - psf_cut_y]
    else:
        return array


def load_imaging_data_from_fits(
    image_path,
    pixel_scale,
    image_hdu=0,
    resized_imaging_shape=None,
    resized_imaging_origin_pixels=None,
    resized_imaging_origin_arcsec=None,
    psf_path=None,
    psf_hdu=0,
    resized_psf_shape=None,
    renormalize_psf=True,
    noise_map_path=None,
    noise_map_hdu=0,
    noise_map_from_image_and_background_noise_map=False,
    convert_noise_map_from_weight_map=False,
    convert_noise_map_from_inverse_noise_map=False,
    background_noise_map_path=None,
    background_noise_map_hdu=0,
    convert_background_noise_map_from_weight_map=False,
    convert_background_noise_map_from_inverse_noise_map=False,
    poisson_noise_map_path=None,
    poisson_noise_map_hdu=0,
    poisson_noise_map_from_image=False,
    convert_poisson_noise_map_from_weight_map=False,
    convert_poisson_noise_map_from_inverse_noise_map=False,
    exposure_time_map_path=None,
    exposure_time_map_hdu=0,
    exposure_time_map_from_single_value=None,
    exposure_time_map_from_inverse_noise_map=False,
    background_sky_map_path=None,
    background_sky_map_hdu=0,
    convert_from_electrons=False,
    gain=None,
    convert_from_adus=False,
    lens_name=None,
):
    """Factory for loading the imaging data_type from .fits files, as well as computing properties like the noise-map,
    exposure-time map, etc. from the imaging-data.

    This factory also includes a number of routines for converting the imaging-data from units not supported by PyAutoLens \
    (e.g. adus, electrons) to electrons per second.

    Parameters
    ----------
    lens_name
    image_path : str
        The path to the image .fits file containing the image (e.g. '/path/to/image.fits')
    pixel_scale : float
        The size of each pixel in arc seconds.
    image_hdu : int
        The hdu the image is contained in the .fits file specified by *image_path*.        
    image_hdu : int
        The hdu the image is contained in the .fits file that *image_path* points too.
    resized_imaging_shape : (int, int) | None
        If input, the imaging arrays that are image sized, e.g. the image, noise-maps) are resized to these dimensions.
    resized_imaging_origin_pixels : (int, int) | None
        If the imaging arrays are resized, this defines a new origin (in pixels) around which recentering occurs.
    resized_imaging_origin_arcsec : (float, float) | None
        If the imaging arrays are resized, this defines a new origin (in arc-seconds) around which recentering occurs.
    psf_path : str
        The path to the psf .fits file containing the psf (e.g. '/path/to/psf.fits')        
    psf_hdu : int
        The hdu the psf is contained in the .fits file specified by *psf_path*.
    resized_psf_shape : (int, int) | None
        If input, the psf is resized to these dimensions.
    renormalize_psf : bool
        If True, the PSF is renoralized such that all elements sum to 1.0.
    noise_map_path : str
        The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')        
    noise_map_hdu : int
        The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
    noise_map_from_image_and_background_noise_map : bool
        If True, the noise-map is computed from the observed image and background noise-map \
        (see NoiseMap.from_image_and_background_noise_map).
    convert_noise_map_from_weight_map : bool
        If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_noise_map_from_inverse_noise_map : bool
        If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
        *NoiseMap.from_inverse_noise_map).
    background_noise_map_path : str
        The path to the background_noise_map .fits file containing the background noise-map \ 
        (e.g. '/path/to/background_noise_map.fits')        
    background_noise_map_hdu : int
        The hdu the background_noise_map is contained in the .fits file specified by *background_noise_map_path*.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    poisson_noise_map_path : str
        The path to the poisson_noise_map .fits file containing the Poisson noise-map \
         (e.g. '/path/to/poisson_noise_map.fits')        
    poisson_noise_map_hdu : int
        The hdu the poisson_noise_map is contained in the .fits file specified by *poisson_noise_map_path*.
    poisson_noise_map_from_image : bool
        If True, the Poisson noise-map is estimated using the image.
    convert_poisson_noise_map_from_weight_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_poisson_noise_map_from_inverse_noise_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    exposure_time_map_path : str
        The path to the exposure_time_map .fits file containing the exposure time map \ 
        (e.g. '/path/to/exposure_time_map.fits')        
    exposure_time_map_hdu : int
        The hdu the exposure_time_map is contained in the .fits file specified by *exposure_time_map_path*.
    exposure_time_map_from_single_value : float
        The exposure time of the imaging, which is used to compute the exposure-time map as a single value \
        (see *ExposureTimeMap.from_single_value*).
    exposure_time_map_from_inverse_noise_map : bool
        If True, the exposure-time map is computed from the background noise_map map \
        (see *ExposureTimeMap.from_background_noise_map*)
    background_sky_map_path : str
        The path to the background_sky_map .fits file containing the background sky map \
        (e.g. '/path/to/background_sky_map.fits').
    background_sky_map_hdu : int
        The hdu the background_sky_map is contained in the .fits file specified by *background_sky_map_path*.
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in units of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in units of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """

    image = abstract_data.load_image(
        image_path=image_path, image_hdu=image_hdu, pixel_scale=pixel_scale
    )

    background_noise_map = load_background_noise_map(
        background_noise_map_path=background_noise_map_path,
        background_noise_map_hdu=background_noise_map_hdu,
        pixel_scale=pixel_scale,
        convert_background_noise_map_from_weight_map=convert_background_noise_map_from_weight_map,
        convert_background_noise_map_from_inverse_noise_map=convert_background_noise_map_from_inverse_noise_map,
    )

    if background_noise_map is not None:
        inverse_noise_map = 1.0 / background_noise_map
    else:
        inverse_noise_map = None

    exposure_time_map = abstract_data.load_exposure_time_map(
        exposure_time_map_path=exposure_time_map_path,
        exposure_time_map_hdu=exposure_time_map_hdu,
        pixel_scale=pixel_scale,
        shape=image.shape,
        exposure_time=exposure_time_map_from_single_value,
        exposure_time_map_from_inverse_noise_map=exposure_time_map_from_inverse_noise_map,
        inverse_noise_map=inverse_noise_map,
    )

    poisson_noise_map = load_poisson_noise_map(
        poisson_noise_map_path=poisson_noise_map_path,
        poisson_noise_map_hdu=poisson_noise_map_hdu,
        pixel_scale=pixel_scale,
        convert_poisson_noise_map_from_weight_map=convert_poisson_noise_map_from_weight_map,
        convert_poisson_noise_map_from_inverse_noise_map=convert_poisson_noise_map_from_inverse_noise_map,
        image=image,
        exposure_time_map=exposure_time_map,
        poisson_noise_map_from_image=poisson_noise_map_from_image,
        convert_from_electrons=convert_from_electrons,
        gain=gain,
        convert_from_adus=convert_from_adus,
    )

    noise_map = load_noise_map(
        noise_map_path=noise_map_path,
        noise_map_hdu=noise_map_hdu,
        pixel_scale=pixel_scale,
        image=image,
        background_noise_map=background_noise_map,
        exposure_time_map=exposure_time_map,
        convert_noise_map_from_weight_map=convert_noise_map_from_weight_map,
        convert_noise_map_from_inverse_noise_map=convert_noise_map_from_inverse_noise_map,
        noise_map_from_image_and_background_noise_map=noise_map_from_image_and_background_noise_map,
        convert_from_electrons=convert_from_electrons,
        gain=gain,
        convert_from_adus=convert_from_adus,
    )

    psf = load_psf(
        psf_path=psf_path,
        psf_hdu=psf_hdu,
        pixel_scale=pixel_scale,
        renormalize=renormalize_psf,
    )

    background_sky_map = load_background_sky_map(
        background_sky_map_path=background_sky_map_path,
        background_sky_map_hdu=background_sky_map_hdu,
        pixel_scale=pixel_scale,
    )

    imaging_data = ImagingData(
        image=image,
        pixel_scale=pixel_scale,
        psf=psf,
        noise_map=noise_map,
        background_noise_map=background_noise_map,
        poisson_noise_map=poisson_noise_map,
        exposure_time_map=exposure_time_map,
        background_sky_map=background_sky_map,
        gain=gain,
        name=lens_name,
    )

    if resized_imaging_shape is not None:
        imaging_data = imaging_data.new_imaging_data_with_resized_arrays(
            new_shape=resized_imaging_shape,
            new_centre_pixels=resized_imaging_origin_pixels,
            new_centre_arcsec=resized_imaging_origin_arcsec,
        )

    if resized_psf_shape is not None:
        imaging_data = imaging_data.new_imaging_data_with_resized_psf(
            new_shape=resized_psf_shape
        )

    if convert_from_electrons:
        imaging_data = imaging_data.new_imaging_data_converted_from_electrons()
    elif convert_from_adus:
        imaging_data = imaging_data.new_imaging_data_converted_from_adus(gain=gain)

    return imaging_data


def load_noise_map(
    noise_map_path,
    noise_map_hdu,
    pixel_scale,
    image=None,
    background_noise_map=None,
    exposure_time_map=None,
    convert_noise_map_from_weight_map=False,
    convert_noise_map_from_inverse_noise_map=False,
    noise_map_from_image_and_background_noise_map=False,
    convert_from_electrons=False,
    gain=None,
    convert_from_adus=False,
):
    """Factory for loading the noise-map from a .fits file.

    This factory also includes a number of routines for converting the noise-map from from other units (e.g. \
    a weight map) or computing the noise-map from other unblurred_image_1d (e.g. the imaging image and background noise-map).

    Parameters
    ----------
    noise_map_path : str
        The path to the noise_map .fits file containing the noise_map (e.g. '/path/to/noise_map.fits')
    noise_map_hdu : int
        The hdu the noise_map is contained in the .fits file specified by *noise_map_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    image : ndarray
        The image-image, which the noise-map can be calculated using.
    background_noise_map : ndarray
        The background noise-map, which the noise-map can be calculated using.
    exposure_time_map : ndarray
        The exposure-time map, which the noise-map can be calculated using.
    convert_noise_map_from_weight_map : bool
        If True, the noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_noise_map_from_inverse_noise_map : bool
        If True, the noise-map loaded from the .fits file is converted from an inverse noise-map to a noise-map (see \
        *NoiseMap.from_inverse_noise_map).
    background_noise_map_path : str
        The path and filename of the .fits image containing the background noise-map.
    background_noise_map_hdu : int
        The hdu the background noise-map is contained in the .fits file that *background_noise_map_path* points too.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    noise_map_from_image_and_background_noise_map : bool
        If True, the noise-map is computed from the observed image and background noise-map \
        (see NoiseMap.from_image_and_background_noise_map).
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in units of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in units of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """
    noise_map_options = sum(
        [
            convert_noise_map_from_weight_map,
            convert_noise_map_from_inverse_noise_map,
            noise_map_from_image_and_background_noise_map,
        ]
    )

    if noise_map_options > 1:
        raise exc.DataException(
            "You have specified more than one method to load the noise_map map, e.g.:"
            "convert_noise_map_from_weight_map | "
            "convert_noise_map_from_inverse_noise_map |"
            "noise_map_from_image_and_background_noise_map"
        )

    if noise_map_options == 0 and noise_map_path is not None:
        return NoiseMap.from_fits_with_pixel_scale(
            file_path=noise_map_path, hdu=noise_map_hdu, pixel_scale=pixel_scale
        )
    elif convert_noise_map_from_weight_map and noise_map_path is not None:
        weight_map = scaled_array.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu
        )
        return NoiseMap.from_weight_map(weight_map=weight_map, pixel_scale=pixel_scale)
    elif convert_noise_map_from_inverse_noise_map and noise_map_path is not None:
        inverse_noise_map = scaled_array.Array.from_fits(
            file_path=noise_map_path, hdu=noise_map_hdu
        )
        return NoiseMap.from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map, pixel_scale=pixel_scale
        )
    elif noise_map_from_image_and_background_noise_map:

        if background_noise_map is None:
            raise exc.DataException(
                "Cannot compute the noise-map from the image and background noise_map map if a "
                "background noise_map map is not supplied."
            )

        if (
            not (convert_from_electrons or convert_from_adus)
            and exposure_time_map is None
        ):
            raise exc.DataException(
                "Cannot compute the noise-map from the image and background noise_map map if an "
                "exposure-time (or exposure time map) is not supplied to convert to adus"
            )

        if convert_from_adus and gain is None:
            raise exc.DataException(
                "Cannot compute the noise-map from the image and background noise_map map if a"
                "gain is not supplied to convert from adus"
            )

        return NoiseMap.from_image_and_background_noise_map(
            pixel_scale=pixel_scale,
            image=image,
            background_noise_map=background_noise_map,
            exposure_time_map=exposure_time_map,
            convert_from_electrons=convert_from_electrons,
            gain=gain,
            convert_from_adus=convert_from_adus,
        )
    else:
        raise exc.DataException(
            "A noise_map map was not loaded, specify a noise_map_path or option to compute a noise_map map."
        )


def load_psf(psf_path, psf_hdu, pixel_scale, renormalize=False):
    """Factory for loading the psf from a .fits file.

    Parameters
    ----------
    psf_path : str
        The path to the psf .fits file containing the psf (e.g. '/path/to/psf.fits')
    psf_hdu : int
        The hdu the psf is contained in the .fits file specified by *psf_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    renormalize : bool
        If True, the PSF is renoralized such that all elements sum to 1.0.
    """
    if renormalize:
        return PSF.from_fits_renormalized(
            file_path=psf_path, hdu=psf_hdu, pixel_scale=pixel_scale
        )
    if not renormalize:
        return PSF.from_fits_with_scale(
            file_path=psf_path, hdu=psf_hdu, pixel_scale=pixel_scale
        )


def load_background_noise_map(
    background_noise_map_path,
    background_noise_map_hdu,
    pixel_scale,
    convert_background_noise_map_from_weight_map=False,
    convert_background_noise_map_from_inverse_noise_map=False,
):
    """Factory for loading the background noise-map from a .fits file.

    This factory also includes a number of routines for converting the background noise-map from from other units (e.g. \
    a weight map).

    Parameters
    ----------
    background_noise_map_path : str
        The path to the background_noise_map .fits file containing the background noise-map \
        (e.g. '/path/to/background_noise_map.fits')
    background_noise_map_hdu : int
        The hdu the background_noise_map is contained in the .fits file specified by *background_noise_map_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    convert_background_noise_map_from_weight_map : bool
        If True, the bacground noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_background_noise_map_from_inverse_noise_map : bool
        If True, the background noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    """
    background_noise_map_options = sum(
        [
            convert_background_noise_map_from_weight_map,
            convert_background_noise_map_from_inverse_noise_map,
        ]
    )

    if background_noise_map_options == 0 and background_noise_map_path is not None:
        return NoiseMap.from_fits_with_pixel_scale(
            file_path=background_noise_map_path,
            hdu=background_noise_map_hdu,
            pixel_scale=pixel_scale,
        )
    elif (
        convert_background_noise_map_from_weight_map
        and background_noise_map_path is not None
    ):
        weight_map = scaled_array.Array.from_fits(
            file_path=background_noise_map_path, hdu=background_noise_map_hdu
        )
        return NoiseMap.from_weight_map(weight_map=weight_map, pixel_scale=pixel_scale)
    elif (
        convert_background_noise_map_from_inverse_noise_map
        and background_noise_map_path is not None
    ):
        inverse_noise_map = scaled_array.Array.from_fits(
            file_path=background_noise_map_path, hdu=background_noise_map_hdu
        )
        return NoiseMap.from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map, pixel_scale=pixel_scale
        )
    else:
        return None


def load_poisson_noise_map(
    poisson_noise_map_path,
    poisson_noise_map_hdu,
    pixel_scale,
    convert_poisson_noise_map_from_weight_map=False,
    convert_poisson_noise_map_from_inverse_noise_map=False,
    poisson_noise_map_from_image=False,
    image=None,
    exposure_time_map=None,
    convert_from_electrons=False,
    gain=None,
    convert_from_adus=False,
):
    """Factory for loading the Poisson noise-map from a .fits file.

    This factory also includes a number of routines for converting the Poisson noise-map from from other units (e.g. \
    a weight map) or computing the Poisson noise_map from other unblurred_image_1d (e.g. the imaging image).

    Parameters
    ----------
    poisson_noise_map_path : str
        The path to the poisson_noise_map .fits file containing the Poisson noise-map \
         (e.g. '/path/to/poisson_noise_map.fits')
    poisson_noise_map_hdu : int
        The hdu the poisson_noise_map is contained in the .fits file specified by *poisson_noise_map_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    convert_poisson_noise_map_from_weight_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from a weight-map to a noise-map (see \
        *NoiseMap.from_weight_map).
    convert_poisson_noise_map_from_inverse_noise_map : bool
        If True, the Poisson noise-map loaded from the .fits file is converted from an inverse noise-map to a \
        noise-map (see *NoiseMap.from_inverse_noise_map).
    poisson_noise_map_from_image : bool
        If True, the Poisson noise-map is estimated using the image.
    image : ndarray
        The image, which the Poisson noise-map can be calculated using.
    background_noise_map : ndarray
        The background noise-map, which the Poisson noise-map can be calculated using.
    exposure_time_map : ndarray
        The exposure-time map, which the Poisson noise-map can be calculated using.
    convert_from_electrons : bool
        If True, the input unblurred_image_1d are in units of electrons and all converted to electrons / second using the exposure \
        time map.
    gain : float
        The image gain, used for convert from ADUs.
    convert_from_adus : bool
        If True, the input unblurred_image_1d are in units of adus and all converted to electrons / second using the exposure \
        time map and gain.
    """
    poisson_noise_map_options = sum(
        [
            convert_poisson_noise_map_from_weight_map,
            convert_poisson_noise_map_from_inverse_noise_map,
            poisson_noise_map_from_image,
        ]
    )

    if poisson_noise_map_options == 0 and poisson_noise_map_path is not None:
        return PoissonNoiseMap.from_fits_with_pixel_scale(
            file_path=poisson_noise_map_path,
            hdu=poisson_noise_map_hdu,
            pixel_scale=pixel_scale,
        )
    elif poisson_noise_map_from_image:

        if (
            not (convert_from_electrons or convert_from_adus)
            and exposure_time_map is None
        ):
            raise exc.DataException(
                "Cannot compute the Poisson noise-map from the image if an "
                "exposure-time (or exposure time map) is not supplied to convert to adus"
            )

        if convert_from_adus and gain is None:
            raise exc.DataException(
                "Cannot compute the Poisson noise-map from the image if a"
                "gain is not supplied to convert from adus"
            )

        return PoissonNoiseMap.from_image_and_exposure_time_map(
            pixel_scale=pixel_scale,
            image=image,
            exposure_time_map=exposure_time_map,
            convert_from_electrons=convert_from_electrons,
            gain=gain,
            convert_from_adus=convert_from_adus,
        )

    elif (
        convert_poisson_noise_map_from_weight_map and poisson_noise_map_path is not None
    ):
        weight_map = scaled_array.Array.from_fits(
            file_path=poisson_noise_map_path, hdu=poisson_noise_map_hdu
        )
        return PoissonNoiseMap.from_weight_map(
            weight_map=weight_map, pixel_scale=pixel_scale
        )
    elif (
        convert_poisson_noise_map_from_inverse_noise_map
        and poisson_noise_map_path is not None
    ):
        inverse_noise_map = scaled_array.Array.from_fits(
            file_path=poisson_noise_map_path, hdu=poisson_noise_map_hdu
        )
        return PoissonNoiseMap.from_inverse_noise_map(
            inverse_noise_map=inverse_noise_map, pixel_scale=pixel_scale
        )
    else:
        return None


def load_background_sky_map(
    background_sky_map_path, background_sky_map_hdu, pixel_scale
):
    """Factory for loading the background sky from a .fits file.

    Parameters
    ----------
    background_sky_map_path : str
        The path to the background_sky_map .fits file containing the background sky map \
        (e.g. '/path/to/background_sky_map.fits').
    background_sky_map_hdu : int
        The hdu the background_sky_map is contained in the .fits file specified by *background_sky_map_path*.
    pixel_scale : float
        The size of each pixel in arc seconds.
    """
    if background_sky_map_path is not None:
        return scaled_array.ScaledSquarePixelArray.from_fits_with_pixel_scale(
            file_path=background_sky_map_path,
            hdu=background_sky_map_hdu,
            pixel_scale=pixel_scale,
        )
    else:
        return None


def output_imaging_data_to_fits(
    imaging_data,
    image_path,
    psf_path,
    noise_map_path=None,
    background_noise_map_path=None,
    poisson_noise_map_path=None,
    exposure_time_map_path=None,
    background_sky_map_path=None,
    overwrite=False,
):
    array_util.numpy_array_2d_to_fits(
        array_2d=imaging_data.image, file_path=image_path, overwrite=overwrite
    )
    array_util.numpy_array_2d_to_fits(
        array_2d=imaging_data.psf, file_path=psf_path, overwrite=overwrite
    )

    if imaging_data.noise_map is not None and noise_map_path is not None:
        array_util.numpy_array_2d_to_fits(
            array_2d=imaging_data.noise_map,
            file_path=noise_map_path,
            overwrite=overwrite,
        )

    if (
        imaging_data.background_noise_map is not None
        and background_noise_map_path is not None
    ):
        array_util.numpy_array_2d_to_fits(
            array_2d=imaging_data.background_noise_map,
            file_path=background_noise_map_path,
            overwrite=overwrite,
        )

    if (
        imaging_data.poisson_noise_map is not None
        and poisson_noise_map_path is not None
    ):
        array_util.numpy_array_2d_to_fits(
            array_2d=imaging_data.poisson_noise_map,
            file_path=poisson_noise_map_path,
            overwrite=overwrite,
        )

    if (
        imaging_data.exposure_time_map is not None
        and exposure_time_map_path is not None
    ):
        array_util.numpy_array_2d_to_fits(
            array_2d=imaging_data.exposure_time_map,
            file_path=exposure_time_map_path,
            overwrite=overwrite,
        )

    if (
        imaging_data.background_sky_map is not None
        and background_sky_map_path is not None
    ):
        array_util.numpy_array_2d_to_fits(
            array_2d=imaging_data.background_sky_map,
            file_path=background_sky_map_path,
            overwrite=overwrite,
        )



from test.integration.tests.full_pipeline import decomposed_lens_light_and_dark
from test.integration.tests.full_pipeline import hyper_no_lens_light_bg
from test.integration.tests.full_pipeline import hyper_with_lens_light
from test.integration.tests.full_pipeline import hyper_with_lens_light_bg
from test.integration.tests.full_pipeline import hyper_with_lens_light_bg_new_api
from test.integration.tests.runner import run_a_mock


class TestCase:
    def _test_decomposed_lens_light_and_dark(self):
        run_a_mock(decomposed_lens_light_and_dark)

    def _test_hyper_no_lens_light_bg(self):
        run_a_mock(hyper_no_lens_light_bg)

    def _test_hyper_with_lens_light(self):
        run_a_mock(hyper_with_lens_light)

    def _test_hyper_with_lens_light_bg(self):
        run_a_mock(hyper_with_lens_light_bg)

    def _test_hyper_with_lens_light_bg_new_api(self):
        run_a_mock(hyper_with_lens_light_bg_new_api)
